import inspect
import typing
import dataclasses

# from types import GenericAlias

MISSING = object()
_FIELDS = getattr(dataclasses, '_FIELDS', '__dataclass_fields__')
_POST_INIT_NAME = getattr(dataclasses, '_POST_INIT_NAME', '__post_init__')


# TODO: add support for InitVar and ClassVar like in original dataclasses


def _set_new_attribute(cls, name, value):
    """
    Never overwrites an existing attribute.  Returns True if the
    attribute already exists.
    """
    if name in cls.__dict__:
        return True
    setattr(cls, name, value)
    return False


def _with_setattr(obj, attribute, value):
    setattr(obj, attribute, value)
    return obj


T = typing.TypeVar('T')


@dataclasses.dataclass
class Box(typing.Generic[T]):
    value: T


class Boxed:
    def __init__(
        self, name=None
    ):  # currently we allways pass name in this persistent.py
        self.name = name

    def __set_name__(self, owner, name):
        if not self.name:
            self.name = name

    def __get__(self, instance, owner):
        if instance is None:
            return self
        return getattr(instance, self.name).value

    def __set__(self, instance, value):
        if not hasattr(instance, self.name):
            if isinstance(value, Box):
                setattr(instance, self.name, value)
            elif inspect.isdatadescriptor(value):
                # try to find default inside descriptor
                default = getattr(value, 'default', None)
                if default is not None:
                    setattr(instance, self.name, Box(default))
                    return
                # try to get default from __get__
                if hasattr(value, '__get__'):
                    try:
                        default = value.__get__(instance, owner=type(instance))
                        if default:
                            setattr(instance, self.name, Box(default))
                            return
                    except Exception:
                        pass
                raise TypeError(
                    f"__init__() missing required argument to be stored as: '{value.name}'"
                    f"or cannot find default value for descriptor: {value} "
                    f"(default should be provided "
                    f"either by 'default' attribute or by __get__ method)"
                )
            else:
                setattr(instance, self.name, Box(value))
        else:
            getattr(instance, self.name).value = value


class Field:
    def __init__(self, name: str, type_: type, default):
        if isinstance(default, (list, dict, set)):
            raise ValueError(
                f"Don't use mutable type {type(default)} as a default for field {name}"
            )

        self.name = name
        self.type_ = type_
        self.default = default

        # for some compatibility with original dataclasses (not used in this impl)
        self.init = True
        self.default_factory = MISSING
        self._field_type = getattr(dataclasses, '_FIELD')

    def __set_name__(self, owner, name):
        if not self.name:
            self.name = name

        # Forward __set_name__ to the field default if it is a descriptor.
        set_name = getattr(type(self.default), '__set_name__', None)
        if set_name:
            # There is a __set_name__ method on the descriptor, call
            # it.
            set_name(self.default, owner, self.box_mask(name))

    @property
    def has_default(self) -> bool:
        return self.default is not MISSING

    @property
    def has_default_as_descriptor(self) -> bool:
        return self.has_default and inspect.isdatadescriptor(self.default)

    @property
    def as_init_arg(self) -> str:
        maybe_default = f'={self.named_default}' if self.has_default else ''
        return f'{self.name}: {self.named_type}{maybe_default}'

    @property
    def as_assignment(self) -> str:
        return f'self.{self.name} = {self.name}'

    @property
    def named_default(self) -> str:
        return f'_dflt_of_{self.name}'

    @property
    def named_type(self) -> str:
        return f'_type_of_{self.name}'

    @staticmethod
    def from_(class_, name: str, type_: type):
        default = getattr(class_, name, MISSING)
        if isinstance(default, Field):
            # supporting attributes with default as field() instance
            a_field = default
            a_field.name = name
            a_field.type_ = type_
        else:
            # – default was passed directly as a value to class attribute...
            # Thus, if default is a descriptor with __set_name__,
            # its name is incorrect and may lead to recursion on usage.
            # To fix this we should mutate default.name.
            # We could do this later,
            # once field instance provides its descriptor on request,
            # when calling field.descriptor...
            # More over, we could not mutate original descriptor,
            # but create a copy or even a new descriptor decorating original one
            # and providing fixed name, while providing delegate access to all other attributes.
            # ...
            # But since we have separate forwarding logic in __set_name__
            # for the case when descriptor is passed explicitly as a field default,
            # lets for less confusion and KISS, for now, just mutate it below:
            a_field = Field(name, type_, default)
            if a_field.has_default_as_descriptor:
                setattr(a_field.default, 'name', a_field.box_name)

        return a_field

    @staticmethod
    def s_from(class_):
        return [
            Field.from_(class_, name, type_)
            for name, type_ in getattr(class_, '__annotations__', {}).items()
        ]

    @staticmethod
    def box_mask(name):
        return f'__boxed_{name}'

    @staticmethod
    def value_from(obj, attribute):
        return getattr(obj, Field.box_mask(attribute)).value

    @property
    def box_name(self):
        return self.box_mask(self.name)

    @property
    def descriptor(self):
        return (
            Boxed(self.box_name)
            if not self.has_default_as_descriptor
            else _with_setattr(self.default, 'name', self.box_name)
        )

    # __class_getitem__ = classmethod(GenericAlias)


# This function is used instead of exposing Field creation directly,
# so that a type checker can be told (via overloads) that this is a
# function whose type depends on its parameters.
def field(
    *,
    default=MISSING,
    # default_factory=MISSING,
    # init=True,
    # repr=True,
    # hash=None,
    # compare=True,
    # metadata=None,
):
    # if default is not MISSING and default_factory is not MISSING:
    #     raise ValueError('cannot specify both default and default_factory')

    return Field(
        name=...,
        type_=...,
        default=default,  # , default_factory, init, repr, hash, compare, metadata
    )


def dataclass(cls):
    """
    Generates __init__ method for the `cls`
    based on type annotations and class attributes with default values.
    Provides a `Boxed` descriptor for each field,
    which persists the reference to the field value
    after object was «cloned» via `replace` method.

    The custom descriptor can be provided:
    * as default value
      * e.g. `nickname: str = ToUpperDescriptor()`
    * or explicit field with default value
      * e.g. `nickname: str = field(default=ToUpperDescriptor())` .
    Then this descriptor will be used instead of `Boxed` descriptor
    to provide get and set access to the field value.
    Take into account, that this custom descriptor
    should handle situation when user forgot to set the required argument
    on object creation (trough implicit call the generated __init__ method).
    Consider sub-classing `Boxed` descriptor for this purpose
    and reuse its __set__ implementation, that raises `TypeError`
    if required argument was missed
    and custom descriptor subclass does not provide the default value
    either by `default` attribute or by `__get__` implementation.

    """
    fields = Field.s_from(cls)
    setattr(cls, _FIELDS, dict((f.name, f) for f in fields))

    locals_ = {}
    init_args = []

    for field in fields:
        locals_[field.named_type] = field.type_
        if field.has_default:
            locals_[field.named_default] = field.default

        init_args.append(field.as_init_arg)

    has_post_init = hasattr(cls, _POST_INIT_NAME)
    post_init_params_str = ','.join(
        field.name
        for field in fields
        if field._field_type is dataclasses._FIELD_INITVAR
    )

    wrapper_fn = '\n'.join(
        [
            f'def wrapper({", ".join(locals_.keys())}):',
            f' def __init__(self, {", ".join(init_args)}):',
            '\n'.join(
                [
                    *[f'  {field.as_assignment}' for field in fields],
                    *(
                        [f'  self.{_POST_INIT_NAME}({post_init_params_str})']
                        if has_post_init
                        else []
                    ),
                ]
            )
            or '  pass',
            ' return __init__',
        ]
    )

    namespace = {}
    # pylint: disable=exec-used
    exec(wrapper_fn, None, namespace)

    init_fn = namespace['wrapper'](**locals_)

    _set_new_attribute(cls, '__init__', init_fn)

    for field in fields:
        setattr(cls, field.name, field.descriptor)

    if not getattr(cls, '__doc__'):
        cls.__doc__ = cls.__name__ + str(inspect.signature(cls)).replace(' -> None', '')

    return cls


def replace(obj, **changes):
    """Return a new object replacing specified fields with new values.

    Example:

        >>> @dataclass
        >>> class C:
        >>>     x: int
        >>>     y: int
        >>>
        >>> c = C(1, 2)
        >>> c1 = replace(c, x=3)
        >>> assert c1.x == 3 and c1.y == 2
    """

    if not dataclasses._is_dataclass_instance(obj):
        raise TypeError('replace() should be called on dataclass instances')

    # If a field is not in 'changes', read its value from the provided obj.
    for f in getattr(obj, _FIELDS).values():
        # Only consider normal fields or InitVars.
        if f._field_type is dataclasses._FIELD_CLASSVAR:
            continue

        if not f.init:
            if f.name in changes:
                raise ValueError(
                    f'field {f.name} is declared with '
                    'init=False, it cannot be specified with '
                    'replace()'
                )
            continue

        if f.name not in changes:
            if f._field_type is dataclasses._FIELD_INITVAR and f.default is MISSING:
                raise ValueError(
                    f"InitVar {f.name!r} " 'must be specified with replace()'
                )
            changes[f.name] = getattr(obj, f.box_name)

    return obj.__class__(**changes)



import abc
import itertools
import typing
from abc import ABC, abstractmethod

from selenium.webdriver.common.options import BaseOptions
from selenium.webdriver.common.service import Service

from selene.common.fp import pipe as pipe, F
from selene.common.helpers import (
    flatten as flatten,
    is_absolute_url as is_absolute_url,
    to_by as to_by,
)
from selene.core.condition import Condition as Condition
from selene.core.configuration import Config as Config
from selene.core.exceptions import TimeoutException as TimeoutException
from selene.core.locator import Locator as Locator
from selene.core.wait import Command as Command, Query as Query, Wait as Wait
from selene.support.webdriver import WebHelper as WebHelper
from selenium.webdriver.remote.switch_to import SwitchTo as SwitchTo
from selenium.webdriver.remote.webdriver import WebDriver as WebDriver
from selenium.webdriver.remote.webelement import WebElement
from typing import Callable, Iterable, Optional, Tuple, TypeVar, Union, Any, Generic

E = TypeVar('E', bound='Assertable')
R = TypeVar('R')

class Assertable(ABC, Generic[E], metaclass=abc.ABCMeta):
    @abstractmethod
    def should(self, condition: Condition[E]) -> E: ...

class Matchable(Assertable[E], metaclass=abc.ABCMeta):
    @abstractmethod
    def wait_until(self, condition: Condition[E]) -> bool: ...
    @abstractmethod
    def matching(self, condition: Condition[E]) -> bool: ...

class Configured(ABC, metaclass=abc.ABCMeta):
    @property
    @abstractmethod
    def config(self) -> Config: ...
    # @abstractmethod
    # def with_(self, config: Optional[Config] = ..., **config_as_kwargs) -> E: ...

class WaitingEntity(Matchable[E], Configured, metaclass=abc.ABCMeta):
    def __init__(self, config: Config) -> None: ...
    @property
    def wait(self) -> Wait[E]: ...
    def perform(self, command: Command[E]) -> E: ...
    def get(self, query: Query[E, R]) -> R: ...
    @property
    def config(self) -> Config: ...
    def should(self, condition: Condition[E]) -> E: ...
    def wait_until(self, condition: Condition[E]) -> bool: ...
    def matching(self, condition: Condition[E]) -> bool: ...

class Element(WaitingEntity['Element']):
    def __init__(self, locator: Locator[WebElement], config: Config) -> None: ...
    def with_(
        self,
        config: Optional[Config] = None,
        *,
        # Options to customize general Selene behavior
        # > to customize waiting logic
        timeout: float = 4,
        # poll_during_waits: int = ...,  # currently fake option
        # _wait_decorator: Callable[[Wait[E]], Callable[[F], F]] = lambda w: lambda f: f,
        # reports_folder: Optional[str] = ...,
        # _counter: itertools.count = ...,
        save_screenshot_on_failure: bool = True,
        save_page_source_on_failure: bool = True,
        # last_screenshot: Optional[str] = None,
        # last_page_source: Optional[str] = None,
        # _save_screenshot_strategy: Callable[[Config, Optional[str]], Any] = ...,
        # _save_page_source_strategy: Callable[[Config, Optional[str]], Any] = ...,
        # # Options to customize web browser and elements behavior
        # base_url: str = '',
        # _get_base_url_on_open_with_no_args: bool = False,
        # window_width: Optional[int] = None,
        # window_height: Optional[int] = None,
        log_outer_html_on_failure: bool = False,
        set_value_by_js: bool = False,
        type_by_js: bool = False,
        click_by_js: bool = False,
        wait_for_no_overlap_found_by_js: bool = False,
        # Etc.
        _build_wait_strategy: Callable[[Config], Callable[[E], Wait[E]]] = ...,
    ) -> Element: ...
    def locate(self) -> WebElement: ...
    @property
    def __raw__(self): ...
    def __call__(self) -> WebElement: ...
    @property
    def wait(self) -> Wait[Element]: ...
    @property
    def cached(self) -> Element: ...
    def element(self, css_or_xpath_or_by: Union[str, Tuple[str, str]]) -> Element: ...
    def all(self, css_or_xpath_or_by: Union[str, Tuple[str, str]]) -> Collection: ...
    def execute_script(self, script_on_self: str, *arguments): ...
    def set_value(self, value: Union[str, int]) -> Element: ...
    def set(self, value: Union[str, int]) -> Element: ...
    def type(self, text: Union[str, int]) -> Element: ...
    def send_keys(self, *value) -> Element: ...
    def press(self, *keys) -> Element: ...
    def press_enter(self) -> Element: ...
    def press_escape(self) -> Element: ...
    def press_tab(self) -> Element: ...
    def clear(self) -> Element: ...
    def submit(self) -> Element: ...
    def click(self) -> Element: ...
    def double_click(self) -> Element: ...
    def context_click(self) -> Element: ...
    def hover(self) -> Element: ...
    def s(self, css_or_xpath_or_by: Union[str, Tuple[str, str]]) -> Element: ...
    def ss(self, css_or_xpath_or_by: Union[str, Tuple[str, str]]) -> Collection: ...

class Collection(WaitingEntity['Collection'], Iterable[Element]):
    def __init__(
        self, locator: Locator[typing.Sequence[WebElement]], config: Config
    ) -> None: ...
    def with_(
        self,
        config: Optional[Config] = None,
        *,
        # Options to customize general Selene behavior
        # > to customize waiting logic
        timeout: float = 4,
        poll_during_waits: int = ...,  # currently fake option
        # _wait_decorator: Callable[[Wait[E]], Callable[[F], F]] = lambda w: lambda f: f,
        # reports_folder: Optional[str] = ...,
        # _counter: itertools.count = ...,
        save_screenshot_on_failure: bool = True,
        save_page_source_on_failure: bool = True,
        # last_screenshot: Optional[str] = None,
        # last_page_source: Optional[str] = None,
        # _save_screenshot_strategy: Callable[[Config, Optional[str]], Any] = ...,
        # _save_page_source_strategy: Callable[[Config, Optional[str]], Any] = ...,
        # # Options to customize web browser and elements behavior
        # base_url: str = '',
        # _get_base_url_on_open_with_no_args: bool = False,
        # window_width: Optional[int] = None,
        # window_height: Optional[int] = None,
        log_outer_html_on_failure: bool = False,
        set_value_by_js: bool = False,
        type_by_js: bool = False,
        click_by_js: bool = False,
        wait_for_no_overlap_found_by_js: bool = False,
        # Etc.
        _build_wait_strategy: Callable[[Config], Callable[[E], Wait[E]]] = ...,
    ) -> Collection: ...
    def locate(self) -> typing.Sequence[WebElement]: ...
    @property
    def __raw__(self): ...
    def __call__(self) -> typing.Sequence[WebElement]: ...
    @property
    def cached(self) -> Collection: ...
    def __iter__(self): ...
    def __len__(self) -> int: ...
    def element(self, index: int) -> Element: ...
    @property
    def first(self) -> Element: ...
    @property
    def second(self) -> Element: ...
    @property
    def even(self): ...
    @property
    def odd(self): ...
    def sliced(
        self, start: Optional[int] = ..., stop: Optional[int] = ..., step: int = ...
    ) -> Collection: ...
    def __getitem__(
        self, index_or_slice: Union[int, slice]
    ) -> Union[Element, Collection]: ...
    def from_(self, start: int) -> Collection: ...
    def to(self, stop: int) -> Collection: ...
    def by(
        self, condition: Union[Condition[Element], Callable[[Element], None]]
    ) -> Collection: ...
    def filtered_by(
        self, condition: Union[Condition[Element], Callable[[Element], None]]
    ) -> Collection: ...
    def by_their(
        self,
        selector: Union[str, Tuple[str, str], Callable[[Element], Element]],
        condition: Condition[Element],
    ) -> Collection: ...
    def element_by(
        self, condition: Union[Condition[Element], Callable[[Element], None]]
    ) -> Element: ...
    def element_by_its(
        self,
        selector: Union[str, Tuple[str, str], Callable[[Element], Element]],
        condition: Condition[Element],
    ) -> Element: ...
    def collected(
        self, finder: Callable[[Element], Union[Element, Collection]]
    ) -> Collection: ...
    def all(self, selector: Union[str, Tuple[str, str]]) -> Collection: ...
    def all_first(self, selector: Union[str, Tuple[str, str]]) -> Collection: ...

class Browser(WaitingEntity['Browser']):
    def __init__(self, config: Optional[Config] = ...) -> None: ...
    def with_(
        self,
        config: Optional[Config] = None,
        *,
        # Options to customize default driver lifecycle
        driver_name: str = 'chrome',
        driver_options: Optional[BaseOptions] = None,
        driver_service: Optional[Service] = None,
        driver_remote_url: Optional[str] = None,
        hold_driver_at_exit: bool = False,
        _reset_not_alive_driver_on_get_url: bool = True,
        rebuild_not_alive_driver: bool = False,
        _driver_get_url_strategy: Callable[
            [Config], Callable[[Optional[str]], None]
        ] = ...,
        # Options to customize driver management
        build_driver_strategy: Callable[[Config], WebDriver] = ...,
        _schedule_driver_teardown_strategy: Callable[
            [Config, Callable[[], WebDriver]], None
        ] = ...,
        _teardown_driver_strategy: Callable[[Config, WebDriver], None] = ...,
        _is_driver_set_strategy: Callable[[WebDriver], bool] = ...,
        _is_driver_alive_strategy: Callable[[WebDriver], bool] = ...,
        # Managed Driver
        driver: WebDriver = ...,
        # Options to customize this config creation
        _override_driver_with_all_driver_like_options: bool = True,
        # Options to customize general Selene behavior
        # > to customize waiting logic
        timeout: float = 4,
        poll_during_waits: int = ...,  # currently fake option
        _wait_decorator: Callable[[Wait[E]], Callable[[F], F]] = lambda w: lambda f: f,
        reports_folder: Optional[str] = ...,
        _counter: itertools.count = ...,
        save_screenshot_on_failure: bool = True,
        save_page_source_on_failure: bool = True,
        last_screenshot: Optional[str] = None,
        last_page_source: Optional[str] = None,
        _save_screenshot_strategy: Callable[[Config, Optional[str]], Any] = ...,
        _save_page_source_strategy: Callable[[Config, Optional[str]], Any] = ...,
        # Options to customize web browser and elements behavior
        base_url: str = '',
        _get_base_url_on_open_with_no_args: bool = False,
        window_width: Optional[int] = None,
        window_height: Optional[int] = None,
        log_outer_html_on_failure: bool = False,
        set_value_by_js: bool = False,
        type_by_js: bool = False,
        click_by_js: bool = False,
        wait_for_no_overlap_found_by_js: bool = False,
        # Etc.
        _build_wait_strategy: Callable[[Config], Callable[[E], Wait[E]]] = ...,
    ) -> Browser: ...
    @property
    def driver(self) -> WebDriver: ...
    @property
    def __raw__(self): ...
    def element(
        self, css_or_xpath_or_by: Union[str, Tuple[str, str], Locator]
    ) -> Element: ...
    def all(
        self, css_or_xpath_or_by: Union[str, Tuple[str, str], Locator]
    ) -> Collection: ...
    def open(self, relative_or_absolute_url: Optional[str] = ...) -> Browser: ...
    def switch_to_next_tab(self) -> Browser: ...
    def switch_to_previous_tab(self) -> Browser: ...
    def switch_to_tab(self, index_or_name: Union[int, str]) -> Browser: ...
    @property
    def switch_to(self) -> SwitchTo: ...
    def quit(self) -> None: ...
    def close(self) -> Browser: ...
    def execute_script(self, script, *args): ...
    def save_screenshot(self, file: Optional[str] = ...): ...
    @property
    def last_screenshot(self) -> str: ...
    def save_page_source(self, file: Optional[str] = ...) -> Optional[str]: ...
    @property
    def last_page_source(self) -> str: ...
    def close_current_tab(self) -> Browser: ...
    def clear_local_storage(self) -> Browser: ...
    def clear_session_storage(self) -> Browser: ...



# MIT License
#
# Copyright (c) 2015-2022 Iakiv Kramarenko
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import os
import typing
import warnings

from abc import abstractmethod, ABC
from typing import TypeVar, Union, Callable, Tuple, Iterable, Optional

from selenium.webdriver.remote.webdriver import WebDriver
from selenium.webdriver import ActionChains
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.remote.switch_to import SwitchTo
from selenium.webdriver.remote.webelement import WebElement

from selene.common.fp import pipe

from selene.core.configuration import Config
from selene.core.wait import Wait, Command, Query
from selene.core.condition import Condition
from selene.core.locator import Locator

from selene.common.helpers import to_by, flatten, is_absolute_url
from selene.core.exceptions import TimeoutException, _SeleneError
from selene.support.webdriver import WebHelper


E = TypeVar('E', bound='Assertable')
R = TypeVar('R')


class Assertable(ABC, typing.Generic[E]):
    @abstractmethod
    # TODO: shouldn't we type self too?
    #       see #generic-methods-and-generic-self
    #       at https://mypy.readthedocs.io/en/stable/generics.html
    # def should(self: E, condition: Condition[E]) -> E:
    def should(self, condition: Condition[E]) -> E:
        pass


# TODO: try as Matchable(ABC) and check if autocomplete will still work
class Matchable(Assertable[E]):
    @abstractmethod
    def wait_until(self, condition: Condition[E]) -> bool:
        pass

    @abstractmethod
    def matching(self, condition: Condition[E]) -> bool:
        pass


# class Configured(ABC, typing.Generic[E]):
class Configured(ABC):
    @property
    @abstractmethod
    def config(self) -> Config:
        pass

    # @abstractmethod
    # def with_(self, config: Optional[Config] = None, **config_as_kwargs) -> E:
    #     pass


class WaitingEntity(Matchable[E], Configured):
    def __init__(self, config: Config):
        self._config = config

    @property
    def wait(self) -> Wait[E]:
        return self.config.wait(typing.cast(E, self))  # type: ignore

    def perform(self, command: Command[E]) -> E:
        """Useful to call external commands.

        Commands might be predefined in Selene:
            element.perform(command.js.scroll_into_view)
        or some custom defined by selene user:
            element.perform(my_action.triple_click)

        You might think that it will be useful
        to use these methods also in Selene internally
        in order to define built in commands e.g. in Element class, like:

            def click(self):
                return self.perform(Command('click', lambda element: element().click()))

        instead of:

            def click(self):
                self.wait.for_(Command('click', lambda element: element().click()))
                return self

        But so far, we use the latter version - though, less concise, but more explicit,
        making it more obvious that waiting is built in;)

        """
        self.wait.for_(command)
        return typing.cast(E, self)

    # TODO: what about `entity[query.something]` syntax
    #       over or in addition to `entity.get(query.something)` ?
    def get(self, query: Query[E, R]) -> R:
        return self.wait.for_(query)

    # --- Configured --- #

    @property
    def config(self) -> Config:
        return self._config

    # --- Assertable --- #

    def should(self, condition: Condition[E]) -> E:
        self.wait.for_(condition)
        return typing.cast(E, self)

    # --- Matchable --- #

    def wait_until(self, condition: Condition[E]) -> bool:
        return self.wait.until(condition)

    def matching(self, condition: Condition[E]) -> bool:
        return condition.predicate(typing.cast(E, self))


class Element(WaitingEntity['Element']):
    @staticmethod
    def _log_webelement_outer_html_for(
        element: Element,
    ) -> Callable[[TimeoutException], Exception]:
        def log_webelement_outer_html(error: TimeoutException) -> Exception:
            from selene.core import query
            from selene.core.match import element_is_present

            cached = element.cached

            if cached.matching(element_is_present):
                return TimeoutException(
                    f'{error.msg}\n'
                    f'Actual webelement: {query.outer_html(element)}'  # type: ignore
                )
            else:
                return error

        return log_webelement_outer_html

    # TODO: should we move locator based init and with_ to Located base abstract class?

    def __init__(self, locator: Locator[WebElement], config: Config):
        self._locator = locator
        super().__init__(config)

    # --- Configured --- #

    def with_(self, config: Optional[Config] = None, **config_as_kwargs) -> Element:
        return Element(
            self._locator,
            config if config else self.config.with_(**config_as_kwargs),
        )

    # --- Located --- #

    def __str__(self):
        return str(self._locator)

    def locate(self) -> WebElement:
        return self._locator()

    @property
    def __raw__(self):
        return self.locate()

    def __call__(self) -> WebElement:
        return self.locate()

    # --- WaitingEntity --- #

    @property
    def wait(self) -> Wait[Element]:
        # TODO:  will not it break code like browser.with_(timeout=...)?
        # TODO: fix that will disable/break shared hooks (snapshots)
        # return Wait(self,  # TODO:  isn't it slower to create it each time from scratch? move to __init__?
        #             at_most=self.config.timeout,
        #             or_fail_with=pipe(
        #                 Element._log_webelement_outer_html_for(self),
        #                 self.config.hook_wait_failure))
        if self.config.log_outer_html_on_failure:
            # TODO: remove this part completely from core.entity logic
            #       move it to support.shared.config
            return super().wait.or_fail_with(
                pipe(
                    Element._log_webelement_outer_html_for(self),
                    super().wait.hook_failure,
                )
            )
        else:
            return super().wait

    @property
    def cached(self) -> Element:
        # TODO: do we need caching ? with lazy save of webelement to cache

        cache = None
        error = None
        try:
            cache = self()
        except Exception as e:
            error = e

        def get_webelement():
            if cache:
                return cache
            raise error

        return Element(Locator(f'{self}.cached', get_webelement), self.config)

    # --- Relative location --- #

    def element(self, css_or_xpath_or_by: Union[str, Tuple[str, str]]) -> Element:
        by = to_by(css_or_xpath_or_by)

        return Element(
            Locator(f'{self}.element({by})', lambda: self().find_element(*by)),
            self.config,
        )

    def all(self, css_or_xpath_or_by: Union[str, Tuple[str, str]]) -> Collection:
        by = to_by(css_or_xpath_or_by)

        return Collection(
            Locator(f'{self}.all({by})', lambda: self().find_elements(*by)),
            self.config,
        )

    # --- Commands --- #

    def execute_script(self, script_on_self: str, *arguments):
        """
        Executes JS script on self as webelement. Will not work for Mobile!

        The script can use predefined parameters:
        - ``element`` and ``self`` are aliases to this element handle, i.e. ``self.locate()`` or ``self()``.
        - ``arguments`` are accessible from the script with same order and indexing as they are provided to the method

        Examples::

            browser.element('[id^=google_ads]').execute_script('element.remove()')
            # OR
            browser.element('[id^=google_ads]').execute_script('self.remove()')
            '''
            # are shortcuts to
            browser.execute_script('arguments[0].remove()', browser.element('[id^=google_ads]')())
            '''

            browser.element('input').execute_script('element.value=arguments[0]', 'new value')
            # OR
            browser.element('input').execute_script('self.value=arguments[0]', 'new value')
            '''
            # are shortcuts to
            browser.execute_script('arguments[0].value=arguments[1]', browser.element('input').locate(), 'new value')
            '''
        """
        driver: WebDriver = self.config.driver
        webelement = self()
        # TODO: should we wrap it in wait or not?
        # TODO: should we add additional it and/or its aliases for element?
        return driver.execute_script(
            f'''
                let element = arguments[0]
                let self = arguments[0]
                return (function(...args) {{
                    {script_on_self}
                }})(...arguments[1])
            ''',
            webelement,
            arguments,
        )

    # TODO: do we need this method?
    #       do we really need to wrap script into function(element,args) here?
    #       if yes... wouldn't it be better to use standard arguments name
    #       instead of args?
    #       for better integration with js support in jetbrains products?
    def _execute_script(
        self,
        script_on_self_element_and_args: str,
        *extra_args,
    ):
        warnings.warn(
            '._execute_script is now deprecated '
            'in favor of .execute_script(script_on_self, *arguments) '
            'that uses access to arguments (NOT args!) in the script',
            DeprecationWarning,
        )
        driver: WebDriver = self.config.driver
        webelement = self()
        # TODO: should we wrap it in wait or not?
        return driver.execute_script(
            f'''
                return (function(element, args) {{
                    {script_on_self_element_and_args}
                }})(arguments[0], arguments[1])
            ''',
            webelement,
            extra_args,
        )

    def set_value(self, value: Union[str, int]) -> Element:
        # TODO: should we move all commands like following or queries like in conditions - to separate py modules?
        # TODO: should we make them webelement based (Callable[[WebElement], None]) instead of element based?
        def fn(element: Element):
            webelement = (
                element._actual_not_overlapped_webelement
                if self.config.wait_for_no_overlap_found_by_js
                else element()
            )
            webelement.clear()  # TODO: change to impl based not on clear, because clear generates post-events...
            webelement.send_keys(str(value))

        from selene.core import command

        # TODO: should we log the webelement source in the command name below?
        #       i.e. change from:
        #
        #   else Command(f'set value: {value}', fn)
        #
        #       to more low level:
        #
        #   else (
        #       Command(f'actual_not_overlapped_webelement.clear().send_keys({value})', fn)
        #       if self.config.wait_for_no_overlap_found_by_js
        #       else
        #       Command(f'actual_webelement.clear().send_keys({value})', fn)
        #   )
        #
        self.wait.for_(
            command.js.set_value(value)
            if self.config.set_value_by_js
            else Command(f'set value: {value}', fn)
        )

        # TODO: consider returning self.cached, since after first successful call,
        #       all next ones should normally pass
        #       no waiting will be needed normally
        #       if yes - then we should pass fn commands to wait.for_ so the latter will return webelement to cache
        #       also it will make sense to make this behaviour configurable...
        return self

    def set(self, value: Union[str, int]) -> Element:
        """
        Sounds similar to Element.set_value(self, value), but considered to be used in broader context.
        For example, a use can say «Set gender radio to Male» but will hardly say «Set gender radio to value Male».
        Now imagine, on your project you have custom html implementation of radio buttons,
        and want to teach selene to set such radio-button controls
        – you can do this via Element.set(self, value) method,
        after monkey-patching it according to your behavior;)
        """
        return self.set_value(value)

    def _actual_visible_webelement_and_maybe_its_cover(
        self, center_x_offset=0, center_y_offset=0
    ) -> Tuple[WebElement, WebElement]:
        # TODO: will it be faster render outerHTML via lazy rendered SeleneError
        #       instead of: throw `element ${element.outerHTML} is not visible`
        #       in below js
        results = self.execute_script(
            '''
                var centerXOffset = arguments[0];
                var centerYOffset = arguments[1];

                var isVisible = !!(
                    element.offsetWidth
                    || element.offsetHeight
                    || element.getClientRects().length
                ) && window.getComputedStyle(element).visibility !== 'hidden'

                if (!isVisible) {
                    throw `element ${element.outerHTML} is not visible`
                }

                var rect = element.getBoundingClientRect();
                var x = rect.left + rect.width/2 + centerXOffset;
                var y = rect.top + rect.height/2 + centerYOffset;

                // TODO: now we return [element, null]
                //       in case of elementFromPoint returns null
                //       (kind of – if we don't know what to do,
                //       let's at least not block the execution...)
                //       rethink this... and handle the iframe case
                //       read more in
// https://developer.mozilla.org/en-US/docs/Web/API/Document/elementFromPoint

                var elementByXnY = document.elementFromPoint(x,y);
                if (elementByXnY == null) {
                    return [element, null];
                }

                var isNotOverlapped = element.isSameNode(elementByXnY);

                return isNotOverlapped
                       ? [element, null]
                       : [element, elementByXnY];
            ''',
            center_x_offset,
            center_y_offset,
        )
        webelement = results[0]
        maybe_cover = results[1]

        return webelement, maybe_cover

    @property
    def _actual_not_overlapped_webelement(self):
        (
            webelement,
            maybe_cover,
        ) = self._actual_visible_webelement_and_maybe_its_cover()
        if maybe_cover is not None:
            raise _SeleneError(
                lambda: f'Element: {webelement.get_attribute("outerHTML")}\n'
                + '\tis overlapped by: '
                + maybe_cover.get_attribute("outerHTML")
            )

        return webelement

    def type(self, text: Union[str, int]) -> Element:
        def fn(element: Element):
            if self.config.wait_for_no_overlap_found_by_js:
                webelement = element._actual_not_overlapped_webelement
            else:
                webelement = element()
            webelement.send_keys(str(text))

        from selene.core import command

        self.wait.for_(
            command.js.type(text)
            if self.config.type_by_js
            else Command(f'type: {text}', fn)
        )

        return self

    def send_keys(self, *value) -> Element:
        """
        To be used for more low level operations like «uploading files», etc.
        To simulate normal input of keys by user when typing
        - use Element.type(self, text).
        """
        self.wait.command('send keys', lambda element: element().send_keys(*value))
        return self

    def press(self, *keys) -> Element:
        def fn(element: Element):
            webelement = (
                element._actual_not_overlapped_webelement
                if self.config.wait_for_no_overlap_found_by_js
                else element()
            )
            webelement.send_keys(*keys)

        self.wait.command(f'press keys: {keys}', fn)

        return self

    def press_enter(self) -> Element:
        return self.press(Keys.ENTER)

    def press_escape(self) -> Element:
        return self.press(Keys.ESCAPE)

    def press_tab(self) -> Element:
        return self.press(Keys.TAB)

    def clear(self) -> Element:
        def fn(element: Element):
            webelement = (
                element._actual_not_overlapped_webelement
                if self.config.wait_for_no_overlap_found_by_js
                else element()
            )
            webelement.clear()

        self.wait.command('clear', fn)

        return self

    def submit(self) -> Element:
        def fn(element: Element):
            webelement = (
                element._actual_not_overlapped_webelement
                if self.config.wait_for_no_overlap_found_by_js
                else element()
            )
            webelement.submit()

        self.wait.command('submit', fn)

        return self

    # TODO: add offset args with defaults, or add additional method, think on what is better
    def click(self) -> Element:
        """Just a normal click:)"""

        def raw_click(element: Element):
            element.locate().click()

        from selene.core import command

        self.wait.for_(
            typing.cast(Command[Element], command.js.click)
            if self.config.click_by_js
            else Command('click', raw_click)
        )

        return self

    def double_click(self) -> Element:
        actions: ActionChains = ActionChains(self.config.driver)

        def fn(element: Element):
            webelement = (
                element._actual_not_overlapped_webelement
                if self.config.wait_for_no_overlap_found_by_js
                else element()
            )
            actions.double_click(webelement).perform()

        self.wait.command('double click', fn)

        return self

    def context_click(self) -> Element:
        actions: ActionChains = ActionChains(self.config.driver)

        def fn(element: Element):
            webelement = (
                element._actual_not_overlapped_webelement
                if self.config.wait_for_no_overlap_found_by_js
                else element()
            )
            actions.context_click(webelement).perform()

        self.wait.command('context click', fn)

        return self

    def hover(self) -> Element:
        actions: ActionChains = ActionChains(self.config.driver)

        def fn(element: Element):
            webelement = (
                element._actual_not_overlapped_webelement
                if self.config.wait_for_no_overlap_found_by_js
                else element()
            )
            actions.move_to_element(webelement).perform()

        self.wait.command('hover', fn)

        return self

    # TODO: should we reflect queries as self methods? or not...
    # pros: faster to query element attributes
    # cons: queries are not test oriented. test is steps + asserts
    #       so queries will be used only occasionally, then why to make a heap from Element?
    #       hence, occasionally it's enough to have them called as
    #           query.outer_html(element)  # non-waiting version
    #       or
    #           element.get(query.outer_html)  # waiting version
    # def outer_html(self) -> str:
    #     return self.wait.for_(query.outer_html)

    # --- Deprecate or not? --- #

    def s(self, css_or_xpath_or_by: Union[str, Tuple[str, str]]) -> Element:
        # warnings.warn(
        #     "consider using more explicit `element` instead: browser.element('#foo').element('.bar')",
        #     SyntaxWarning)
        return self.element(css_or_xpath_or_by)

    def ss(self, css_or_xpath_or_by: Union[str, Tuple[str, str]]) -> Collection:
        # warnings.warn(
        #     "consider using `all` instead: browser.element('#foo').all('.bar')",
        #     SyntaxWarning)
        return self.all(css_or_xpath_or_by)


class Collection(WaitingEntity['Collection'], Iterable[Element]):
    def __init__(self, locator: Locator[typing.Sequence[WebElement]], config: Config):
        self._locator = locator
        super().__init__(config)

    def with_(self, config: Optional[Config] = None, **config_as_kwargs) -> Collection:
        return Collection(
            self._locator,
            config if config else self.config.with_(**config_as_kwargs),
        )

    def __str__(self):
        return str(self._locator)

    def locate(self) -> typing.Sequence[WebElement]:
        return self._locator()

    @property
    def __raw__(self):
        return self.locate()

    def __call__(self) -> typing.Sequence[WebElement]:
        return self.locate()

    @property
    def cached(self) -> Collection:
        webelements = self()
        return Collection(Locator(f'{self}.cached', lambda: webelements), self.config)

    def __iter__(self):
        i = 0
        cached = self.cached
        while i < len(cached()):
            element = cached[i]
            yield element
            i += 1

    def __len__(self):
        from selene.core import query

        return self.get(query.size)

    # TODO: add config.index_collection_from_1, disabled by default
    def element(self, index: int) -> Element:
        def find() -> WebElement:
            webelements = self.locate()
            length = len(webelements)

            if length <= index:
                raise AssertionError(
                    f'Cannot get element with index {index} '
                    + f'from webelements collection with length {length}'
                )

            return webelements[index]

        return Element(Locator(f'{self}[{index}]', find), self.config)

    @property
    def first(self) -> Element:
        """
        A human-readable alias to .element(0) or [0]
        """
        return typing.cast(Element, self[0])

    @property
    def second(self) -> Element:
        """
        A human-readable alias to .element(1) or [1]
        """
        return typing.cast(Element, self[1])

    @property
    def even(self):
        """
        A human-readable alias to [1::2], i.e. filtering collection to have only even elements
        """
        return self[1::2]

    @property
    def odd(self):
        """
        A human-readable alias to [::2], i.e. filtering collection to have only odd elements
        """
        return self[::2]

    def sliced(
        self,
        start: Optional[int] = None,
        stop: Optional[int] = None,
        step: int = 1,
    ) -> Collection:
        def find() -> typing.Sequence[WebElement]:
            webelements = self()

            # TODO: assert length according to provided start, stop...

            return webelements[start:stop:step]

        return Collection(
            Locator(
                f'{self}[{start or ""}'
                f':{stop or ""}'
                f'{":" + str(step) if step else ""}]',
                find,
            ),
            self.config,
        )

    def __getitem__(
        self, index_or_slice: Union[int, slice]
    ) -> Union[Element, Collection]:
        if isinstance(index_or_slice, slice):
            return self.sliced(
                index_or_slice.start, index_or_slice.stop, index_or_slice.step
            )

        return self.element(index_or_slice)

    def from_(self, start: int) -> Collection:
        return typing.cast(Collection, self[start:])

    def to(self, stop: int) -> Collection:
        return typing.cast(Collection, self[:stop])

    def by(
        self, condition: Union[Condition[Element], Callable[[Element], None]]
    ) -> Collection:
        condition = (
            condition
            if isinstance(condition, Condition)
            else Condition(str(condition), condition)
        )

        return Collection(
            Locator(
                f'{self}.filtered_by({condition})',
                lambda: [
                    element() for element in self.cached if element.matching(condition)
                ],
            ),
            self.config,
        )

    def filtered_by(
        self, condition: Union[Condition[Element], Callable[[Element], None]]
    ) -> Collection:
        warnings.warn(
            'collection.filtered_by(condition) is deprecated in favor of collection.by(condition)',
            DeprecationWarning,
        )
        return self.by(condition)

    def by_their(
        self,
        selector: Union[str, Tuple[str, str], Callable[[Element], Element]],
        condition: Condition[Element],
    ) -> Collection:
        """
        Returns elements from collection that have inner/relative element,
        found by ``selector`` and matching ``condition``.

        Is a shortcut for ``collection.by(lambda element: condition(element.element(selector))``.

        Example (straightforward)
        -------------------------

        GIVEN html elements somewhere in DOM::
            .result
                .result-title
                .result-url
                .result-snippet

        THEN::

            browser.all('.result')\
                .by_their('.result-title', have.text('Selene'))\
                .should(have.size(3))

        is similar to::

            browser.all('.result')\
                .by_their(lambda it: have.text(text)(it.element('.result-title')))\
                .should(have.size(3))

        Example (PageObject)
        --------------------

        GIVEN html elements somewhere in DOM::
            .result
                .result-title
                .result-url
                .result-snippet

        AND::

            results = browser.all('.result')
            class Result:
                def __init__(self, element):
                    self.element = element
                    self.title = self.element.element('.result-title')
                    self.url = self.element.element('.result-url')
            # ...

        THEN::

            results.by_their(lambda it: Result(it).title, have.text(text))\
                .should(have.size(3))

        is similar to::

            results.by_their(lambda it: have.text(text)(Result(it).title))\
                .should(have.size(3))
        """

        def find_in(parent: Element) -> Element:
            if callable(selector):
                return selector(parent)
            else:
                return parent.element(selector)

        return self.by(lambda it: condition(find_in(it)))

    def element_by(
        self, condition: Union[Condition[Element], Callable[[Element], None]]
    ) -> Element:
        # TODO: a first_by(condition) alias would be shorter,
        #  and more consistent with by(condition).first
        #  but the phrase items.element_by(have.text('foo')) leads to a more
        #  natural meaning that such element should be only one...
        #  while items.first_by(have.text('foo')) gives a clue that
        #  it's just one of many...
        #  should we then make element_by fail
        #  if the condition matches more than one element? (maybe we can control it via corresponding config option?)
        #  yet we don't fail if browser.element(selector) or element.element(selector)
        #  finds more than one element... o_O

        # TODO: In the implementation below...
        #       We use condition in context of "matching", i.e. as a predicate...
        #       why then not accept Callable[[E], bool] also?
        #       (as you remember, Condition is Callable[[E], None] throwing Error)
        #       This will allow the following code be possible
        #           results.element_by(lambda it:
        #               Result(it).title.matching(have.text(text)))
        #       instead of:
        #           results.element_by(lambda it: have.text(text)(
        #                              Result(it).title))
        #       in addition to:
        #           results.element_by_its(lambda it:
        #               Result(it).title, have.text(text))
        #       Open Points:
        #       - do we need element_by_its, if we allow Callable[[E], bool] ?
        #       - if we add elements_by_its, do we need then to accept Callable[[E], bool] ?
        #       - probably... Callable[[E], bool] will lead to worse error messages,
        #         in such case we ignore thrown error's message
        #         - hm... ut seems like we nevertheless ignore it...
        #           we use element.matching(condition) below
        condition = (
            condition
            if isinstance(condition, Condition)
            else Condition(str(condition), condition)
        )

        def find() -> WebElement:
            cached = self.cached

            for element in cached:
                if element.matching(condition):
                    return element()

            from selene.core import query

            if self.config.log_outer_html_on_failure:
                """
                TODO: move it support.shared.config
                """
                outer_htmls = [query.outer_html(element) for element in cached]

                raise AssertionError(
                    f'\n\tCannot find element by condition «{condition}» '
                    f'\n\tAmong {self}'
                    f'\n\tActual webelements collection:'
                    f'\n\t{outer_htmls}'
                )  # TODO: isn't it better to print it all the time via hook, like for Element?
            else:
                raise AssertionError(
                    f'\n\tCannot find element by condition «{condition}» '
                    f'\n\tAmong {self}'
                )

        return Element(Locator(f'{self}.element_by({condition})', find), self.config)

    def element_by_its(
        self,
        selector: Union[str, Tuple[str, str], Callable[[Element], Element]],
        condition: Condition[Element],
    ) -> Element:
        """
        Returns element from collection that has inner/relative element
        found by ``selector`` and matching ``condition``.
        Is a shortcut for ``collection.element_by(lambda its: condition(its.element(selector))``.

        Example (straightforward)
        -------------------------

        GIVEN html elements somewhere in DOM::

            .result
                .result-title
                .result-url
                .result-snippet

        THEN::

            browser.all('.result')\
                .element_by_its('.result-title', have.text(text))\
                .element('.result-url').click()

        ... is a shortcut for::

            browser.all('.result')\
                .element_by(lambda its: have.text(text)(its.element('.result-title')))\
                .element('.result-url').click()

        Example (PageObject)
        --------------------

        GIVEN html elements somewhere in DOM::

            .result
                .result-title
                .result-url
                .result-snippet

        AND::

            results = browser.all('.result')
            class Result:
                def __init__(self, element):
                    self.element = element
                    self.title = self.element.element('.result-title')
                    self.url = self.element.element('.result-url')

        THEN::

            Result(results.element_by_its(lambda it: Result(it).title, have.text(text)))\
                .url.click()

        is a shortcut for::

            Result(results.element_by(lambda it: have.text(text)(Result(it).title)))\
                .url.click()
            # ...
        """

        # TODO: tune implementation to ensure error messages are ok

        def find_in(parent: Element):
            if callable(selector):
                return selector(parent)
            else:
                return parent.element(selector)

        return self.element_by(lambda it: condition(find_in(it)))

    def collected(
        self, finder: Callable[[Element], Union[Element, Collection]]
    ) -> Collection:
        # TODO: consider adding predefined queries to be able to write
        #         collected(query.element(selector))
        #       over
        #         collected(lambda element: element.element(selector))
        #       and
        #         collected(query.all(selector))
        #       over
        #         collected(lambda element: element.all(selector))
        #       consider also putting such element builders like to find.* module instead of query.* module
        #       because they are not supposed to be used in entity.get(*) context defined for other query.* fns

        return Collection(
            Locator(
                f'{self}.collected({finder})',
                # TODO: consider skipping None while flattening
                lambda: typing.cast(
                    typing.Sequence[WebElement],
                    flatten([finder(element)() for element in self.cached]),
                ),
            ),
            self.config,
        )

    def all(self, selector: Union[str, Tuple[str, str]]) -> Collection:
        """
        Returns a collection of all elements found be selector inside each element of self

        An alias to ``collection.collected(lambda its: its.all(selector))``.

        Example
        -------

        Given html::

            <table>
              <tr class="row">
                <td class="cell">A1</td><td class="cell">A2</td>
              </tr>
              <tr class="row">
                <td class="cell">B1</td><td class="cell">B2</td>
              </tr>
            </table>

        Then::

            browser.all('.row').all('.cell')).should(have.texts('A1', 'A2', 'B1', 'B2'))
        """
        by = to_by(selector)

        # TODO: consider implement it through calling self.collected
        #       because actually the impl is self.collected(lambda element: element.all(selector))

        return Collection(
            Locator(
                f'{self}.all({by})',
                lambda: typing.cast(
                    typing.Sequence[WebElement],
                    flatten([webelement.find_elements(*by) for webelement in self()]),
                ),
            ),
            self.config,
        )

    def all_first(self, selector: Union[str, Tuple[str, str]]) -> Collection:
        """
        Returns a collection of each first element found be selector inside each element of self

        An alias to ``collection.collected(lambda its: its.element(selector))``.
        Not same as ``collection.all(selector).first`` that is same as ``collection.first.element(selector)``

        Example
        -------

        Given html::

            <table>
              <tr class="row">
                <td class="cell">A1</td><td class="cell">A2</td>
              </tr>
              <tr class="row">
                <td class="cell">B1</td><td class="cell">B2</td>
              </tr>
            </table>

        Then::

            browser.all('.row').all_first('.cell')).should(have.texts('A1', 'B1'))
        """
        by = to_by(selector)

        # TODO: consider implement it through calling self.collected
        #       because actually the impl is self.collected(lambda element: element.element(selector))

        return Collection(
            Locator(
                f'{self}.all_first({by})',
                lambda: [webelement.find_element(*by) for webelement in self()],
            ),
            self.config,
        )


class Browser(WaitingEntity['Browser']):
    def __init__(self, config: Optional[Config] = None):
        config = Config() if config is None else config
        super().__init__(config)

    def with_(self, config: Optional[Config] = None, **config_as_kwargs) -> Browser:
        return (
            Browser(config)
            if config
            else Browser(self.config.with_(**config_as_kwargs))
        )

    def __str__(self):
        return 'browser'

    @property
    def driver(self) -> WebDriver:
        return self.config.driver

    # TODO: consider making it callable (self.__call__() to be shortcut to self.__raw__ ...)

    @property
    def __raw__(self):
        return self.config.driver

    # @property
    # def actions(self) -> ActionChains:
    #     """
    #     It's kind of just a shortcut for pretty low level actions from selenium webdriver
    #     Yet unsure about this property here:)
    #     comparing to usual high level Selene API...
    #     Maybe later it would be better to make own Actions with selene-style retries, etc.
    #     """
    #     return ActionChains(self.config.driver)

    # --- Element builders --- #

    # TODO: consider None by default,
    #       and *args, **kwargs to be able to pass custom things
    #       to be processed by config.location_strategy
    #       and by default process none as "element to skip all actions on it"
    def element(
        self, css_or_xpath_or_by: Union[str, Tuple[str, str], Locator]
    ) -> Element:
        if isinstance(css_or_xpath_or_by, Locator):
            return Element(css_or_xpath_or_by, self.config)

        by = to_by(css_or_xpath_or_by)

        return Element(
            Locator(f'{self}.element({by})', lambda: self.driver.find_element(*by)),
            self.config,
        )

    def all(
        self, css_or_xpath_or_by: Union[str, Tuple[str, str], Locator]
    ) -> Collection:
        if isinstance(css_or_xpath_or_by, Locator):
            return Collection(css_or_xpath_or_by, self.config)

        by = to_by(css_or_xpath_or_by)

        return Collection(
            Locator(f'{self}.all({by})', lambda: self.driver.find_elements(*by)),
            self.config,
        )

    # --- High Level Commands--- #

    def open(self, relative_or_absolute_url: Optional[str] = None) -> Browser:
        # TODO: should we keep it less pretty but more KISS? like:
        # self.config._driver_get_url_strategy(self.config)(relative_or_absolute_url)
        self.config._executor.get_url(relative_or_absolute_url)

        return self

    def switch_to_next_tab(self) -> Browser:
        from selene.core import query

        self.driver.switch_to.window(query.next_tab(self))

        # TODO: should we use waiting version here (and in other similar cases)?
        # self.perform(Command(
        #     'open next tab',
        #     lambda browser: browser.driver.switch_to.window(query.next_tab(self))))

        return self

    def switch_to_previous_tab(self) -> Browser:
        from selene.core import query

        self.driver.switch_to.window(query.previous_tab(self))
        return self

    def switch_to_tab(self, index_or_name: Union[int, str]) -> Browser:
        if isinstance(index_or_name, int):
            index = index_or_name
            from selene.core import query

            self.driver.switch_to.window(query.tab(index)(self))
        else:
            self.driver.switch_to.window(index_or_name)

        return self

    # TODO: consider deprecating
    @property
    def switch_to(self) -> SwitchTo:
        return self.driver.switch_to

    # TODO: should we add also a shortcut for self.driver.switch_to.alert ?
    #       if we don't need to switch_to.'back' after switch to alert - then for sure we should...
    #       question is - should we implement our own alert as waiting entity?

    def quit(self) -> None:
        """
        Quits the driver.

        If the driver was not even set, will build it just to quit it:D.

        Will fail if the driver was already quit or crashed.
        """
        self.driver.quit()

    # TODO: consider deprecating, it does not close browser, it closes current tab/window
    def close(self) -> Browser:
        self.driver.close()
        return self

    # --- Deprecated --- #

    # TODO: should we keep it?
    def execute_script(self, script, *args):
        warnings.warn(
            'consider using browser.driver.execute_script '
            'instead of browser.execute_script',
            PendingDeprecationWarning,
        )
        return self.driver.execute_script(script, *args)

    # TODO: should we move it to query.* and/or command.*?
    #       like `browser.get(query.screenshot)` ?
    #       like `browser.perform(command.save_screenshot)` ?
    # TODO: deprecate file name, use path
    #       because we can path folder path not file path and it will work
    def save_screenshot(self, file: Optional[str] = None):
        warnings.warn(
            'browser.save_screenshot is deprecated, '
            'use browser.get(query.screenshot_saved())',
            DeprecationWarning,
        )

        from selene.core import query  # type: ignore

        return self.get(query.screenshot_saved())  # type: ignore

    @property
    def last_screenshot(self) -> str:
        warnings.warn(
            'browser.last_screenshot is deprecated, '
            'use browser.config.last_screenshot',
            DeprecationWarning,
        )
        return self.config.last_screenshot  # type: ignore

    # TODO: consider moving this to browser command.save_page_source(filename)
    def save_page_source(self, file: Optional[str] = None) -> Optional[str]:
        warnings.warn(
            'browser.save_page_source is deprecated, '
            'use browser.config.last_screenshot',
            DeprecationWarning,
        )

        if file is None:
            file = self.config._generate_filename(suffix='.html')  # type: ignore

        saved_file = WebHelper(self.driver).save_page_source(file)

        self.config.last_page_source = saved_file  # type: ignore

        return saved_file

    @property
    def last_page_source(self) -> str:
        warnings.warn(
            'browser.last_page_source is deprecated, '
            'use browser.config.last_page_source',
            DeprecationWarning,
        )
        return self.config.last_page_source  # type: ignore

    def close_current_tab(self) -> Browser:
        warnings.warn(
            'deprecated because the «tab» term is not relevant for mobile; '
            'use a `browser.close()` or `browser.driver.close()` instead',
            DeprecationWarning,
        )
        self.driver.close()
        return self

    def clear_local_storage(self) -> Browser:
        warnings.warn(
            'deprecated because of js nature and not-relevance for mobile; '
            'use `browser.perform(command.js.clear_local_storage)` instead',
            DeprecationWarning,
        )
        from selene.core import command

        self.perform(command.js.clear_local_storage)
        return self

    def clear_session_storage(self) -> Browser:
        warnings.warn(
            'deprecated because of js nature and not-relevance for mobile; '
            'use `browser.perform(command.js.clear_session_storage)` instead',
            DeprecationWarning,
        )
        from selene.core import command

        self.perform(command.js.clear_session_storage)
        return self



# MIT License
#
# Copyright (c) 2015-2022 Iakiv Kramarenko
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import functools
from typing import TypeVar, Callable, Any, Tuple, Optional

# T = TypeVar('T', bound=Callable[..., Any])
T = TypeVar('T')
'''
A generic TypeVar to identify a Function Type, i.e. a function
but to be imported and used as fp.T so you can guess from full name,
that it is "function type", i.e. a function ;)
'''
F = TypeVar('F')
'''
A generic TypeVar to identify a Function Type, i.e. a function
can be imported and used as F directly, no need of `fp.F` version...,
you still can guess a shortcut F = Function)
'''

_Decorator = Callable[[F], F]
_Decorator.__doc__ = """
    A type to represent a simple python decorator,
    defined here to type hint arguments and vars in a more readable way,
    so users can guess meaning from IDE autocompletion/other-hints...
    But it happened that e.g. PyCharm will yet render _Decorator simply to (F) -> F.
    Though, something like:
    # class Decorator(Callable[[F], F]):
    #     ...
    would be rendered like _Decorator...
    But is it even proper way in python to define types for type hints? (i.e. "subclassing" other types?)
    """

'''
# TODO: should we have something like?
_DecoratorFactory = Callable[..., Callable[[F], F]]
# pylint: disable=W2301
_DecoratorWithOptionalParameters = Callable[[Optional[F], ...], F]
# the last one even will not work for py3.7 :(
'''


def identity(it: T) -> T:
    return it


def pipe(*functions) -> Optional[Callable[[Any], Any]]:
    """
    pipes functions one by one in the provided order
    i.e. applies arg1, then arg2, then arg3, and so on
    if any arg is None, just skips it
    """
    return (
        functools.reduce(
            lambda f, g: lambda x: f(g(x)) if g else f(x),
            functions[::-1],
            lambda x: x,
        )
        if functions
        else None
    )


def thread(arg, *functions):
    return pipe(*functions)(arg)


def do(function: Callable[[T], Any]) -> Callable[[T], T]:
    def func(arg: T) -> T:
        function(arg)
        return arg

    return func


def write_silently(
    file: str, string: str, encoding: str = 'utf-8'
) -> Optional[Tuple[str, str]]:
    try:
        with open(file, 'w', encoding=encoding) as f:
            f.write(string)
            return file, string
    except OSError:
        return None



# MIT License
#
# Copyright (c) 2015-2023 Iakiv Kramarenko
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

import atexit
import inspect
import itertools
import os
import time
import typing
import warnings
from typing import Callable, Optional, Any

from selenium.webdriver.common.options import BaseOptions
from selenium.webdriver.common.service import Service

from selene.common import fp, helpers
from selene.common.data_structures import persistent
from selene.common.fp import F
from selene.common.helpers import on_error_return_false

from selene.core.exceptions import TimeoutException
from selenium.webdriver.remote.webdriver import WebDriver

from selene.core.wait import Wait, E


# TODO: consider moving to support.*
#       like support._loging.wait_with
def _build_local_driver_by_name_or_remote_by_url_and_options(
    config: Config,
) -> WebDriver:
    from selenium.webdriver import (
        ChromeOptions,
        EdgeOptions,
        Chrome,
        Firefox,
        Edge,
    )

    from selenium.webdriver.chrome.service import Service as ChromeService
    from selenium.webdriver.firefox.service import Service as FirefoxService
    from selenium.webdriver.edge.service import Service as EdgeService  # type: ignore

    from webdriver_manager.chrome import ChromeDriverManager  # type: ignore
    from webdriver_manager.firefox import GeckoDriverManager  # type: ignore
    from webdriver_manager.microsoft import EdgeChromiumDriverManager  # type: ignore

    from webdriver_manager.core.utils import ChromeType  # type: ignore

    def install_and_build_chrome():
        # TODO: consider simplifying the logic... to much of ifs
        #       probably all ifs were already before calling this function
        #       see example of simplification in install_and_build_firefox
        if config.driver_options:
            if isinstance(config.driver_options, ChromeOptions):
                return Chrome(
                    service=config.driver_service
                    or ChromeService(
                        ChromeDriverManager(chrome_type=ChromeType.GOOGLE).install()
                    ),
                    options=config.driver_options,
                )
            else:
                raise ValueError(
                    f'Default config.build_driver_strategy («driver factory»), '
                    f'if config.driver_name is set to "chrome", - '
                    f'expects only instance of ChromeOptions or None'
                    f'in config.driver_options,'
                    f'but got: {config.driver_options}'
                )
        else:
            return Chrome(
                service=config.driver_service
                or ChromeService(
                    ChromeDriverManager(chrome_type=ChromeType.GOOGLE).install()
                )
            )

    def install_and_build_firefox():
        return Firefox(
            service=config.driver_service
            or FirefoxService(GeckoDriverManager().install()),
            options=config.driver_options,
        )

    def install_and_build_edge():
        if config.driver_options:
            if isinstance(config.driver_options, EdgeOptions):
                return Edge(
                    service=config.driver_service
                    or EdgeService(EdgeChromiumDriverManager().install()),
                    options=config.driver_options,
                )
            else:
                raise ValueError(
                    f'Default config.build_driver_strategy, '
                    f'if config.driver_name is set to "edge", - '
                    f'expects only instance of EdgeOptions or None in config.driver_options,'
                    f'but got: {config.driver_options}'
                )
        else:
            return Edge(
                service=config.driver_service or (EdgeChromiumDriverManager().install())
            )

    def build_remote_driver():
        from selenium.webdriver import Remote

        # TODO: consider guessing browserstack remote url
        #       if noticed 'bstack:options' in config.driver_options

        return Remote(
            command_executor=config.driver_remote_url,
            options=config.driver_options,
        )

    def build_appium_driver():
        try:
            from appium import webdriver
        except ImportError as error:
            raise ImportError(
                'Appium-Python-Client is not installed, '
                'run `pip install Appium-Python-Client`,'
                'or add and install dependency '
                'with your favorite dependency manager like poetry: '
                '`poetry add Appium-Python-Client`'
            ) from error

        # TODO: consider to add more smart guessing of options if not set...
        #       like if driver_name is set to 'appium'
        #       and driver_options is not set
        #       and the base_url is set to url of some web app
        #       then build appium driver options
        #       to run web test on mobile browser
        #       else if base_url is set to app path or url,
        #       parse app type and build corresponding appium driver options
        #       ...
        #       TODO: should we even rename base_url to app_url
        #             to cover both web and mobile? or just app?
        #             what about keeping both?
        #             but allowing to set only one of them at same moment?

        return webdriver.Remote(
            command_executor=(
                config.driver_remote_url
                if config.driver_remote_url
                else 'http://127.0.0.1:4723/wd/hub'
            ),
            options=config.driver_options,
        )

    return {  # type: ignore
        'chrome': install_and_build_chrome,
        'firefox': install_and_build_firefox,
        'edge': install_and_build_edge,
        'remote': build_remote_driver,
        'appium': build_appium_driver,
    }.get(
        'appium'
        if (
            config.driver_name == 'appium'
            or (
                config.driver_options
                and 'platformName' in config.driver_options.capabilities
                and config.driver_options.capabilities['platformName'].lower()
                in ['android', 'ios']
            )
        )
        else 'remote'
        if (config.driver_remote_url or config.driver_name == 'remote')
        # TODO: consider automatically detect installed browser if driver_name not set
        else (
            config.driver_name
            if config.driver_name
            else config.driver_options.capabilities['browserName']
            if (
                config.driver_options
                and 'browserName' in config.driver_options.capabilities
            )
            else 'chrome'
        )
    )()


def _maybe_reset_driver_then_tune_window_and_get_with_base_url(config: Config):
    def get(url: Optional[str] = None) -> None:
        if (
            config._reset_not_alive_driver_on_get_url
            and config._executor.is_driver_set
            and config._executor.is_driver_managed
            and not config._executor.is_driver_alive
        ):
            # TODO: consider logging this reset
            #       so user will see it and decide to disable it
            config.driver = typing.cast(WebDriver, ...)

        driver = config.driver

        relative_or_absolute_url = url
        if relative_or_absolute_url is None:
            # force to init driver and open browser or app (for mobile)
            # _ = config.driver  # TODO: why not doing this in all cases?
            if not config.base_url:
                # do nothing more
                return
            if not config._get_base_url_on_open_with_no_args:
                # yet do nothing more
                return
            # proceed with adjusted relative url
            # to be concatenated with base url
            relative_or_absolute_url = ''

        # TODO: skip for mobile
        width = config.window_width
        height = config.window_height

        if width or height:
            if not (width and height):
                size = driver.get_window_size()
                width = width or size['width']
                height = height or size['height']

            driver.set_window_size(int(width), int(height))

        is_absolute = helpers.is_absolute_url(relative_or_absolute_url)
        base_url = config.base_url
        url = (
            relative_or_absolute_url
            if is_absolute
            else base_url + relative_or_absolute_url
        )

        # TODO: should we wrap it into wait? at least for logging?
        driver.get(url)

    return get


# TODO: should we do a complete Manager from it, not just executor,
#       by moving the corresponding logic from _ManagedDriverDescriptor?
# TODO: do we even need it? shouldn't we keep it more KISS?
class _DriverStrategiesExecutor:
    def __init__(self, config: Config):
        self.config = config

    @property
    def driver_instance(self) -> typing.Union[Optional[WebDriver], ...]:  # type: ignore
        return persistent.Field.value_from(self.config, 'driver')

    @property
    def is_driver_managed(self) -> bool:
        return not callable(self.driver_instance)

    def build_driver(self) -> WebDriver:
        return self.config.build_driver_strategy(self.config)

    # TODO: is_set or is_driver_set?
    #       kind of depends on class name.
    #       If named as ManagedDriver, then is_set
    #       If named as DriverManager, then is_driver_set
    @property
    def is_driver_set(self) -> bool:
        return self.config._is_driver_set_strategy(self.driver_instance)

    @property
    def is_driver_alive(self) -> bool:
        # TODO: should we check first if driver is set here?
        return self.config._is_driver_alive_strategy(self.driver_instance)  # type: ignore

    # TODO: this property suits better
    #       for something called as DriverManager not ManagedDriver ;)
    @property
    def teardown(self) -> Callable[[WebDriver], None]:
        return self.config._teardown_driver_strategy(self.config)

    # TODO: this property suits better
    #       for something called as DriverManager not ManagedDriver ;)
    def schedule_teardown(self, get_driver: Callable[[], WebDriver]) -> None:
        self.config._schedule_driver_teardown_strategy(self.config, get_driver)

    def get_url(self, url: Optional[str] = None) -> None:
        self.config._driver_get_url_strategy(self.config)(url)

    def save_screenshot(self, path: Optional[str] = None) -> Any:
        return self.config._save_screenshot_strategy(self.config, path)

    def save_page_source(self, path: Optional[str] = None) -> Any:
        return self.config._save_page_source_strategy(self.config, path)


# TODO: consider reusing config._executor inside this descriptor
class _ManagedDriverDescriptor:
    def __init__(
        self, *, default: typing.Union[Optional[WebDriver], ...] = ...  # type: ignore
    ):
        self.default = default
        self.name = None

    def __set_name__(self, owner, name):
        self.name = name

    def __get__(self, instance, owner):
        if instance is None:
            return self

        config = typing.cast(Config, instance)
        # Below...
        # we can't access driver via config.driver explicitly
        # or implicitly by calling other config.* methods,
        # because it will lead to recursion!!!

        driver_box = typing.cast(persistent.Box[WebDriver], getattr(config, self.name))
        if (
            driver_box.value is None
            or driver_box.value is ...
            or (
                # TODO: think on: if turned on, may slow down tests...
                #       especially when running remote tests...
                config.rebuild_not_alive_driver
                and not callable(driver_box.value)  # TODO: consider deprecating
                and not config._is_driver_alive_strategy(driver_box.value)
            )
        ):
            driver = config.build_driver_strategy(config)
            driver_box.value = driver
            config._schedule_driver_teardown_strategy(config, lambda: driver)

        value = driver_box.value
        if callable(value):
            # warnings.warn(
            #     'Providing driver as callable might be deprecated in future. '
            #     'Consider customizing driver management '
            #     'via other config.* options',
            #     FutureWarning,
            # )
            return value()

        return value

    def __set__(self, instance, value):
        config = typing.cast(Config, instance)

        if not hasattr(instance, self.name):
            # setting this attribute for the first time,
            # probably (TODO: probably or for sure?) in the __init__ method

            if isinstance(value, persistent.Box):
                # it's a boxed value,
                # probably passed implicitly via `persistent.replace`
                driver_box = value
            elif inspect.isdatadescriptor(value):
                # the value happened to be a descriptor
                # it's either object of this descriptor type (`type(self)`)
                # or custom provided descriptor during init new object
                if type(value) is type(self):
                    # we are processing this `self` descriptor as value
                    # so, instead of value, we should store `self.default`
                    driver_box = persistent.Box(self.default)
                else:
                    # somebody decided to provide his own descriptor object
                    # Heh:) It was a good try, but no ;P
                    raise TypeError(
                        'Providing custom descriptor objects on init '
                        'to customize driver management is not supported, '
                        'because it would be too limited... '
                        'You would be able to provide it only on init,'
                        'and use it only via attribute access,'
                        'without possibility to override value with `persistent.replace` '
                        'or `config.with_(**optioins_to_override)`. '
                        'If you want to use custom descriptor, '
                        'you have to subclass Config and provide your descriptor object'
                        'on class attributes definition level.'
                    )  # TODO: cover with tests
            else:
                # setting WebDriver instance directly on init
                driver_box = persistent.Box(value)

                # TODO: here we could remember somehow that driver was set manually
                #       do we need this?

                # currently passing driver as callable disables driver teardown
                if not callable(value):
                    config._schedule_driver_teardown_strategy(config, lambda: value)

            setattr(instance, self.name, driver_box)
        else:
            # setting WebDriver instance after init
            driver_box = getattr(instance, self.name)
            driver_box.value = value

            # currently passing driver as callable disables driver teardown
            if not callable(value):
                # TODO: should not we check value set above,
                #       wasn't the same as was in driver_box.value before?
                #       if yes, we might not want to add one more atexit handler
                config._schedule_driver_teardown_strategy(config, lambda: value)


@persistent.dataclass
class Config:
    """
    A one cross-cutting-concern-like object to group all options
    that might influence Selene behavior depending on context.
    For example, `config.timeout` is used in all "waiting" logic
    of Selene commands. And `config.base_url` is used
    in `browser.open(relative_url)` command.

    As option, the driver instance is also considered. Moreover, this config
    is not just config, but fully manages the driver lifecycle.
    Actually, the "driver manager" is a part of this config.

    While surfing through all avaialable options, pay attention to terminology:

    - all options that have a `driver` word in their name
      are related to driver management, and they are connected in a specific way:)
      - read more on this under `config.with_` doc section;)
    - all options that have a `strategy` word in their name directly influence
      the driver lifecycle in context of driver management.
    - all options that are prefixed with `_` are considered "experimental"
        - their naming can be changed in the future, or even an option can be removed

    Examples:
        Here's how you can build a driver with the instance of this config:

        >>> from selene import Config
        >>> config = Config()
        >>> driver = config.driver  # new instance, built on 1st access to `driver`
        >>> assert driver.name == 'chrome'

        Or pre-configuring the firefox driver:

        >>> from selene import Config
        >>> config = Config(driver_name='firefox')
        >>> driver = config.driver
        >>> assert driver.name == 'firefox'

        Or post-configuring the firefox driver:

        >>> from selene import Config
        >>> config = Config()
        >>> config.driver_name = 'firefox'
        >>> driver = config.driver
        >>> assert driver.name == 'firefox'

        Selene has already predefined shared instance of Config,
        so you can economize on lines of code;)...

        >>> from selene.support.shared import config
        >>> config.driver_name = 'firefox'
        >>> driver = config.driver
        >>> assert driver.name == 'firefox'

        Same shared Config instance is available as browser.config:

        >>> from selene import browser
        >>> browser.config.driver_name = 'firefox'
        >>> driver = browser.config.driver
        >>> assert driver.name == 'firefox'

        There is an alternative style of customizing config.
        The `config.option_name = value` is known in programming
        as "imperative programming" style. When you are creating
        a new Config from scratch, you are actually using
        a "declarative programming" style:

        >>> from selene import Config
        >>> my_config = Config(driver_name='firefox')
        >>> driver = my_config.driver
        >>> assert driver.name == 'firefox'

        Here is an alternative declarative style of
        customizing new config by copying existing:

        >>> from selene import browser
        >>> my_config = browser.config.with_(driver_name='firefox')
        >>> driver = my_config.driver
        >>> assert driver.name == 'firefox'
        >>> # AND...
        >>> assert driver is not browser.config.driver  # ;)
        >>> assert browser.config.driver.name == 'chrome'

        As you can see Selene config is closely related to the browser.
        Moreover, the same type of "declarative config copying" happens implicitely,
        when you apply "copying" to browser:

        >>> from selene import browser
        >>> second_browser = browser.with_(driver_name='firefox')
        >>> assert second_browser.config.driver.name == 'firefox'
        >>> # AND...
        >>> assert second_browser.config.driver is not browser.config.driver  # ;)
        >>> assert browser.config.driver.name == 'chrome'

        Moreover, if you need only a driver, you can have it
        via `browser.driver` shortcut, thus, completely hiding the config:

        >>> from selene import browser
        >>> second_browser = browser.with_(driver_name='firefox')
        >>> assert second_browser.driver.name == 'firefox'
        >>> # AND...
        >>> assert second_browser.driver is not browser.driver  # ;)
        >>> assert browser.driver.name == 'chrome'

        - such shortcut exists only for the `driver` option of config,
        not for other options like `timeout` or `base_url`.
        More nuances of `browser` behavior find in its docs;).

        And here are some more examples of customizing config
        for common test automation use cases...

        Scenario: "Run locally in headless Chrome"

        >>> from selene import browser
        >>> from selenium import webdriver
        >>>
        >>> options = webdriver.ChromeOptions()
        >>> options.add_argument('--headless')
        >>> # additional options:
        >>> options.add_argument('--no-sandbox')
        >>> options.add_argument('--disable-gpu')
        >>> options.add_argument('--disable-notifications')
        >>> options.add_argument('--disable-extensions')
        >>> options.add_argument('--disable-infobars')
        >>> options.add_argument('--enable-automation')
        >>> options.add_argument('--disable-dev-shm-usage')
        >>> options.add_argument('--disable-setuid-sandbox')
        >>> browser.config.driver_options = options

        Scenario: "Run remotely on Selenoid"

        >>> import os
        >>> from selene import browser
        >>> from selenium import webdriver
        >>>
        >>> options = webdriver.ChromeOptions()
        >>> options.browser_version = '100.0'
        >>> options.set_capability(
        >>>     'selenoid:options',
        >>>     {
        >>>         'screenResolution': '1920x1080x24',
        >>>         'enableVNC': True,
        >>>         'enableVideo': True,
        >>>         'enableLog': True,
        >>>     },
        >>> )
        >>> browser.config.driver_options = options
        >>> browser.config.driver_remote_url = (
        >>>     f'https://{os.getenv("LOGIN")}:{os.getenv("PASSWORD")}@'
        >>>     f'selenoid.autotests.cloud/wd/hub'
        >>> )

        Scenario: "Run remotely on BrowserStack in iOS Safari"

        >>> import os
        >>> from selene import browser
        >>> from selenium.webdriver.common.options import ArgOptions
        >>>
        >>> options = ArgOptions()
        >>> options.set_capability(
        >>>     'bstack:options',
        >>>     {
        >>>         'deviceName': 'iPhone 14 Pro Max',
        >>>         # 'browserName': 'safari',  # default for iPhone
        >>>         'userName': os.getenv('BROWSERSTACK_USERNAME'),
        >>>         'accessKey': os.getenv('BROWSERSTACK_ACCESS_KEY'),
        >>>     },
        >>> )
        >>> browser.config.driver_options = options
        >>> browser.config.driver_remote_url = 'http://hub.browserstack.com/wd/hub'

        Scenario: "Run locally in Android Chrome"

        >>> from selene import browser
        >>> from appium.options.common import AppiumOptions
        >>>
        >>> mobile_options = AppiumOptions()
        >>> mobile_options.new_command_timeout = 60
        >>> # Mandatory, also tells Selene to build Appium driver:
        >>> mobile_options.platform_name = 'android'
        >>> mobile_options.set_capability('browserName', 'chrome')
        >>>
        >>> browser.config.driver_options = mobile_options
        >>> # Not mandatory, because it is the default value:
        >>> # browser.config.driver_remote_url = 'http://127.0.0.1:4723/wd/hub'

        Scenario: "Run locally in Android App"

        >>> from selene import browser
        >>> from appium.options.android import UiAutomator2Options
        >>>
        >>> android_options = UiAutomator2Options()
        >>> android_options.new_command_timeout = 60
        >>> android_options.app = 'wikipedia-alpha-universal-release.apk'
        >>> android_options.app_wait_activity = 'org.wikipedia.*'
        >>>
        >>> browser.config.driver_options = android_options

        Scenario: "Run remotely in Android App on BrowserStack"

        >>> import os
        >>> from selene import browser
        >>> from appium.options.android import UiAutomator2Options
        >>>
        >>> options = UiAutomator2Options()
        >>> options.app = 'bs://c700ce60cf13ae8ed97705a55b8e022f13c5827c'
        >>> options.set_capability(
        >>>     'bstack:options',
        >>>     {
        >>>         'deviceName': 'Google Pixel 7',
        >>>         'userName': os.getenv('BROWSERSTACK_USERNAME'),
        >>>         'accessKey': os.getenv('BROWSERSTACK_ACCESS_KEY'),
        >>>     },
        >>> )
        >>> browser.config.driver_options = options
        >>> browser.config.driver_remote_url = 'http://hub.browserstack.com/wd/hub'

    By having config options that influences Selene behavior,
    like `config.timeout` and `config.base_url`,
    - together with complete "driver management",
    we definitely break SRP principle... In the name of Good:D. Kind of;).

    All this makes it far from being a simple options data class...
    - yet kept as one «class for everything» to keep things easier to use,
    especially taking into account some historical reasons of Selenes design,
    that was influenced a lot by the Selenide from Java world.
    As a result sometimes options are not consistent with each other,
    when we speak about different contexts of their usage.
    For example, this same config,
    once customized with `config.driver_options = UiAutomator2Options()`,
    will result in mobile driver built, but then all other web-related options,
    for example, a `config.base_url` will be not relevant.
    Some of them will be ignored, while some of them,
    for example js-related, like `config.set_value_by_js`,
    will break the code execution (JavaScript does not work in mobile apps).
    In an ideal world, we would have to split this config into several ones,
    starting BaseConfig and continuing with WebConfig, MobileConfig, etc.
    Yet, we have what we have. This complicates things a bit,
    especially for us, contributors of Selene,
    but makes easier for newbies in a lot of "harder" cases,
    like customizing same shared browser instance for multi-platform test runs,
    when we have one test that works for all platforms.
    Thus, we allow to do "harder" tasks easier for "less experienced" users.
    Again, such "easiness" does not mean "simplicity" for us, contributors,
    and also for advanced Selene users,
    who want to customize things in a lot of ways
    and have them easier to support on a long run.
    But for now, let's keep it as is, considered as a trade-off.
    """

    # TODO: should it be more as a strategy builder (i.e. curried function)?
    #       i.e.
    #           build_driver_strategy: Callable[
    #             [Config], Callable[[], WebDriver]
    #           ]
    #       for consistency with some other options...
    build_driver_strategy: Callable[
        [Config], WebDriver
    ] = _build_local_driver_by_name_or_remote_by_url_and_options
    """
    A factory to build a driver instance based on this config instance.
    The driver built with this factory will be available via `config.driver`.
    Hence, you can't use `config.driver` directly inside this factory,
    because it may lead to recursion.

    The default factory builds:

    - either a local driver by value specified in `config.driver_name`
    - or a local driver by browserName capability specified in `config.driver_options`
    - or remote driver by value specified in `config.driver_remote_url`.
    - or mobile driver according to `config.driver_options` capabilities.
    """

    # TODO: isn't this option too much?
    #       having it, we have to keep driver descriptor definition
    #       after this option definition,
    #       that is pretty tightly coupled...
    #       heh, but maybe we definitely have to keep it defined
    #       after all "strategy" options...
    # Currently we don't use the power of get_driver being callable...
    # It would work even if we pass simply driver instance...
    # Should we simplify things? Or keep it as is with get_driver?
    _schedule_driver_teardown_strategy: Callable[
        [Config, Callable[[], WebDriver]],
        typing.Union[None, typing.Any],
    ] = lambda config, get_driver: atexit.register(
        lambda: config._teardown_driver_strategy(config)(get_driver())
    )
    """
    Defines when drier will be teardown.
    Is supposed to use config._teardown_driver_strategy under the hood.

    By default, it's registered as an atexit handler.
    """

    # TODO: since it's curried, shouldn't we rename it driver_teardown_strategy?
    _teardown_driver_strategy: Callable[
        [Config], Callable[[WebDriver], None]
    ] = lambda config: lambda driver: (
        driver.quit()
        if not config.hold_driver_at_exit
        and config._is_driver_set_strategy(driver)
        and config._is_driver_alive_strategy(driver)
        else None
    )
    """
    Defines how driver will be teardown.

    By default it quits the driver if it's alive and not asked to be held at exit
    via `config.hold_driver_at_exit`.
    """

    # TODO: should we make it private so far?
    # TODO: shouldn't it be config-based?
    _is_driver_set_strategy: Callable[[Optional[WebDriver]], bool] = lambda driver: (
        driver is not ... and driver is not None
    )
    """
    Defines how to check if driver is set, and so defines how to "unset" or "reset" it.
    """

    # TODO: should we make it private so far?
    _is_driver_alive_strategy: Callable[[WebDriver], bool] = lambda driver: (
        # on_error_return_false(lambda: driver.title is not None)
        (driver.service.process is not None and driver.service.process.poll() is None)
        if hasattr(driver, 'service')
        else on_error_return_false(lambda: driver.window_handles is not None)
    )
    """
    Defines the logic of checking driver for being alive.

    Is supposed to be used in context of triggering automatic driver rebuild,
    depending on context.
    """

    driver_options: Optional[BaseOptions] = None
    """
    Individual browser options to be used on building a driver.

    Examples:
        Can be used instead of `config.driver_name` to tell Selene
        which driver to build, e.g. just specifying

        >>> from selene import browser
        >>> from selenium import webdriver
        >>>
        >>> browser.config.driver_options = webdriver.FirefoxOptions()`

        - will tell Selene to build a Firefox driver.

        But usually you want something more specific,
        like specifying to run a browser in headless more:

        >>> from selene import browser
        >>> from selenium import webdriver
        >>>
        >>> options = webdriver.ChromeOptions()
        >>> options.add_argument('--headless')
        >>> browser.config.driver_options = options
    """

    driver_service: Optional[Service] = None
    """
    Service instance for managing the starting and stopping of the driver.
    Might be useful, for example, for customizing driver executable path,
    if you want to use a custom driver executable instead of the one,
    downloaded by Selene automatically via WDM.
    """

    # Probably, more precise and technically correct name and signature would be:
    #     driver_remote_connection: Optional[Union[str, RemoteConnection]] = None
    # But we decided to keep it more simple and user-friendly
    # in context of the majority of use cases when we just need to pass a URL:
    # for appium and remote cases
    driver_remote_url: Optional[str] = None
    """
    A URL to be used as remote server address to instantiate a RemoteConnection
    to be used by RemoteWebDriver to connect to the remote server.

    Also known as `command_executor`,
    when passing on init: `driver = remote.WebDriver(command_executor=HERE)`.
    Currently we name it and type hint as URL,
    but if you pass a RemoteConnection object,
    it will work same way as in Selenium WebDriver.
    """

    # TODO: consider typing as Optional[Literal['chrome', 'firefox', 'edge', 'appium']]
    # TODO: consider setting to None or ... by default,
    #       and pick up by factory any installed browser in a system
    driver_name: Optional[str] = None
    """
    A desired name of the driver to build by `config.build_driver_strategy`.

    If not set (i.e. set to None, that is a current default value),
    the 'chrome' driver will be used by default.

    It is ignored by default `config.build_driver_strategy`
    if `config.driver_remote_url` is set.

    If you are going to provide your desired driver options
    via `config.driver_options`,
    then Selene will try to guess the corresponding driver name
    based on the options you provided. I.e. no need to provide both:

    ```python
    config.driver_name = 'chrome'
    config.driver_options = ChromeOptions()
    ```

    It's enough to provide only the options:

    ```python
    config.driver_options = ChromeOptions()
    ```

    GIVEN set to any of: 'chrome', 'firefox', 'edge',<br>
    AND config.driver is left unset (default value is ...),<br>
    THEN default config.build_driver_strategy will automatically install drivers<br>
    AND build webdriver instance for you<br>
    AND this config will store the instance in config.driver<br>
    """

    # TODO: finalize the name of this option and consider making public
    _override_driver_with_all_driver_like_options: bool = True
    """
    Controls whether driver will be deep copied
    with `config.driver_name`, `config.driver_remote_url`,
    and so for any other `config.*driver*` option.
    when customizing config via `config.with_(**options_to_override)`.

    Examples:
        Building 2 drivers with implicit deep copy of driver storage:

        >>> chrome_config = Config(
        >>>     driver_name='chrome',
        >>>     timeout=10.0,
        >>>     base_url='https://autotest.how',
        >>> )
        >>> chrome = chrome_config.driver
        >>> firefox_config = chrome_config.with_(driver_name='firefox')
        >>> firefox = firefox_config.driver
        >>> assert firefox is not chrome

        Building 2 drivers with explicit deep copy of driver storage [1]:

        >>> chrome_config = Config(
        >>>     driver_name='chrome',
        >>>     timeout=10.0,
        >>>     base_url='https://autotest.how',
        >>>     _override_driver_with_all_driver_like_options=False,
        >>> )
        >>> chrome = chrome_config.driver
        >>> firefox_config = chrome_config.with_(driver_name='firefox', driver=...)
        >>> firefox = firefox_config.driver
        >>> assert firefox is not chrome

        Building 2 drivers with explicit deep copy of driver storage [2]:

        >>> chrome_config = Config(
        >>>     driver_name='chrome',
        >>>     timeout=10.0,
        >>>     base_url='https://autotest.how',
        >>> )
        >>> chrome_config._override_driver_with_all_driver_like_options = False
        >>> chrome = chrome_config.driver
        >>> firefox_config = chrome_config.with_(name='firefox', driver=...)
        >>> firefox = firefox_config.driver
        >>> assert firefox is not chrome

        Building 1 driver because driver storage was not copied:

        >>> chrome_config = Config(
        >>>     driver_name='chrome',
        >>>     timeout=10.0,
        >>>     base_url='https://autotest.how',
        >>> )
        >>> chrome_config._override_driver_with_all_driver_like_options = False
        >>> chrome = chrome_config.driver
        >>> firefox_config = chrome_config.with_(name='firefox')
        >>> firefox = firefox_config.driver
        >>> assert firefox is chrome  # o_O ;)
    """

    # TODO: consider to deprecate because might confuse in case of Appium usage
    @property
    def browser_name(self) -> Optional[str]:
        return self.driver_name

    # TODO: consider to deprecate because might confuse in case of Appium usage
    @browser_name.setter
    def browser_name(self, value: str):
        self.driver_name = value

    # TODO: do we need it?
    # quit_last_driver_on_reset: bool = False
    # """Controls whether driver will be automatically quit at reset of config.driver"""

    hold_driver_at_exit: bool = False
    """
    Controls whether driver will be automatically quit at process exit or not.

    Will not take much effect on Chrome
    for 4.5.0 < selenium versions <= 4.8.3 < ?.?.?,
    Because for some reason, Selenium of such versions kills driver by himself,
    regardless of what Selene thinks about it:D
    """

    @property
    def hold_browser_open(self) -> bool:
        warnings.warn(
            'Was deprecated because "browser" term '
            'is not relevant to mobile context. '
            'Use `config.hold_driver_at_exit` instead',
            DeprecationWarning,
        )
        return self.hold_driver_at_exit

    @hold_browser_open.setter
    def hold_browser_open(self, value: bool):
        warnings.warn(
            'Was deprecated because "browser" term '
            'is not relevant to mobile context. '
            'Use `config.hold_driver_at_exit = ...` instead',
            DeprecationWarning,
        )
        self.hold_driver_at_exit = value

    # TODO: maybe like this:
    #         _driver_get_url_strategy: Callable[[Config, str], None] = get_driver
    #       or should we implement same "decorator-like" style
    #       for other config-based strategies too?
    # TODO: refactor to inline definition with lambda style
    _driver_get_url_strategy: Callable[
        [Config],
        Callable[[Optional[str]], None],
    ] = _maybe_reset_driver_then_tune_window_and_get_with_base_url
    """
    Defines how to get url with driver depending on other options, like config.base_url.

    Is used inside `browser.open(relative_or_absolute_url)`,
    and defines its behavior correspondingly.
    """

    # TODO: should we use `rebuild` term instead of `reset`?
    #       to be consistent with `rebuild_not_alive_driver`...
    #       Technically, we are not explicitely rebuilding it with this option,
    #       we do reset it by setting to `...`. But then in same `get_url`,
    #       on first access it will be rebuilt automatically.
    _reset_not_alive_driver_on_get_url: bool = True
    """
    Controls whether driver should be automatically reset and, thus,
    forced to be rebuilt, if it was noticed as not alive (e.g. after quit or crash)
    on next call to `config.driver.get(url)`
    (via `config._driver_get_url_strategy`).

    Does not work if `config.driver` was set manually to `Callable[[], WebDriver]`.

    Is a more "week" option than `config.rebuild_not_alive_driver`,
    that is disabled by default,
    that forces to rebuild driver on any next access.
    """

    rebuild_not_alive_driver: bool = False
    """
    Controls whether driver should be automatically rebuilt
    when on next call to config.driver
    it was noticed as not alive (e.g. after quit or crash).

    May slow down your tests if running against remote Selenium server,
    e.g. Grid or selenoid, because of additional request to check
    if driver is alive per each driver action.

    Does not work if `config.driver` was set manually to `Callable[[], WebDriver]`.

    Is a more "strong" option than `config._reset_not_alive_driver_on_get_url`,
    (enabled by default), that schedules rebuilding driver
    on next access only inside "get url" logic.
    """

    # # TODO: why it is not working as attribute + post init?
    # _executor: _DriverStrategiesExecutor = typing.cast(_DriverStrategiesExecutor, ...)
    # def __post_init__(self):
    #     self._executor = _DriverStrategiesExecutor(self)
    @property
    def _executor(self):
        return _DriverStrategiesExecutor(self)

    # TODO: consider allowing to provide a Descriptor as a value
    #       by inheritance like:
    #           class MyConfig(Config):
    #              driver: WebDriver = HERE_DriverDescriptor(...)
    #       and then `MyConfig(driver=...)` will work as expected
    # TODO: should we accept a callable here to bypass build_driver_strategy logic?
    #       currently we do... we don't show it explicitly...
    #       but the valid type is Union[WebDriver, Callable[[], WebDriver]]
    #       so... should we do it?
    #       why not just use config.build_driver_strategy for same?
    #       there is the difference though...
    #       the driver factory is only used as a driver builder,
    #       it does not cover other stages of driver lifecycle,
    #       like teardown...
    #       but if we provide a callable instance to driver,
    #       then it will just substitute the whole lifecycle
    # TODO: consider renaming the class of *Descriptor.
    #       It can be named as ManagedDriverDescriptor, because this is what it is.
    #       But since we have another object named as _managed_driver,
    #       it might be confusing... so maybe DriverManagerDescriptor?
    #       or DriverLifeCycleDescriptor?
    #       or is it ok to have similar names? We still have a Descriptor suffix...
    driver: WebDriver = _ManagedDriverDescriptor(default=...)  # type: ignore
    """
    A driver instance with lifecycle managed by this config special options
    depending on their values and customization of this attribute.

    Once driver-definition-related options are set
    (like `config.driver_options`, `config.driver_remote_url`),
    the driver will be built on first access to this attribute.
    Thus, to build the driver with Selene,
    you simply call `config.driver` for the first time,
    and usually the simplest way to access it
    - is through either `browser.config.driver` or even `browser.driver` shortcut.
    Moreover, usually you don't do this explicitly,
    but rather use `browser.open(url)` to build a driver and open a page.

    Scenarios:
        GIVEN unset, i.e. equals to default `...` or `None`,<br>
        WHEN accessed first time (e.g. via config.driver)<br>
        THEN it will be set to the instance built by `config.build_driver_strategy`.<br>

        GIVEN set manually to an existing driver instance,
              like: `config.driver = Chrome()`<br>
        THEN it will be reused as it is on any next access<br>
        WHEN reset to `...` OR `None`<br>
        THEN will be rebuilt by `config.build_driver_strategy`<br>

        GIVEN set manually to an existing driver instance (not callable),
              like: `config.driver = Chrome()`<br>
        AND at some point of time the driver is not alive
            like crashed or quit<br>
        AND `config._reset_not_alive_driver_on_get_url` is set to `True`,
            that is default<br>
        WHEN driver.get(url) is requested under the hood
             like at `browser.open(url)`<br>
        THEN config.driver will be reset to `...`<br>
        AND thus will be rebuilt by `config.build_driver_strategy`<br>

        GIVEN set manually to a callable that returns WebDriver instance
              (currently marked with FutureWarning, so might be deprecated)<br>
        WHEN accessed fist time<br>
        AND any next time<br>
        THEN will call the callable and return the result<br>

        GIVEN unset or set manually to not callable<br>
        AND `config.hold_driver_at_exit` is set to `False` (that is default)<br>
        WHEN the process exits<br>
        THEN driver will be quit.<br>
    """

    timeout: float = 4
    """
    A default timeout for all Selene waiting that happens under the hood
    of the majority of Selene commands and assertions.
    """

    poll_during_waits: int = 100
    """
    A fake option, not currently used in Selene waiting:)
    """

    # --- Web-specific options ---
    # TODO: should we pass here None?
    #       and use "not None" as _get_base_url_on_open_with_no_args=True?
    # TODO: should we rename it to app_url? or even just app?
    #       and use it as app capability for mobile?
    #       if not set in driver_options...
    base_url: str = ''
    """
    A base url to be used when opening a page with relative url.

    Examples:
        Instead of duplicating the same base url in all your tests:

        >>> from selene import browser
        >>> browser.open('https://mywebapp.xyz/signin')
        >>> ...
        >>> browser.open('https://mywebapp.xyz/signup')
        >>> ...
        >>> browser.open('https://mywebapp.xyz/profile')
        >>> ...

        You can set it once in your config and then just use relative urls:

        >>> from selene import browser
        >>> browser.config.base_url = 'https://mywebapp.xyz'
        >>> browser.open('/signin')
        >>> ...
        >>> browser.open('/signup')
        >>> ...
        >>> browser.open('/profile')
        >>> ...
    """
    # TODO: when adding driver_get_url_strategy
    #       should we rename it to get_base_url_when_relative_url_is_missed?
    #       should we use driver term in the name?
    _get_base_url_on_open_with_no_args: bool = False
    """
    A flag to indicate whether to use `config.base_url`
    when opening a page with `browser.open()` command without arguments.

    If you call `browser.open()` without arguments,
    it will just force the driver to be built and real browser to be opened.

    Even if you have set `config.base_url`, to reuse it in your tests on `browser.open`,
    you have to specify the url explicitly, like `browser.open('/')`
    or at least `browser.open('')`.

    But there are cases where you would like to load base url on `browser.open()`,
    for example in context of cross-platform testing. Then you use this option;)
    See example at
    https://github.com/yashaka/selene/blob/master/examples/run_cross_platform
    """
    window_width: Optional[int] = None
    """
    If set, will be used to set the window width on next call to `browser.open(url)`.
    """
    window_height: Optional[int] = None
    """
    If set, will be used to set the window height on next call to `browser.open(url)`.
    """
    log_outer_html_on_failure: bool = False
    """
    If set to True, will log outer html of the element on failure of any Selene command.

    Is disabled by default, because:

    - it might add too much of noise to the logs
    - will not work on mobile app tests because under the hood - uses JavaScript
    """
    set_value_by_js: bool = False
    """
    A flag to indicate whether to use JavaScript to set value of an element
    on `element.set_value(value)` for purposes of speeding up the test execution,
    or as a workaround when default selenium-based implementation does not work.
    """
    type_by_js: bool = False
    """
    A flag to indicate whether to use JavaScript to type text to an element
    on `element.type(text)` for purposes of speeding up the test execution,
    or as a workaround when default selenium-based implementation does not work.
    """
    click_by_js: bool = False
    """
    A flag to indicate whether to use JavaScript to click on element
    via `element.click()`, usually, as a workaround,
    when default selenium-based implementation does not work.
    """
    wait_for_no_overlap_found_by_js: bool = False
    """
    A flag to indicate whether to use JavaScript to detect overlapping elements
    and wait for them to disappear, when calling commands like `element.type(text)`.
    It is needed because Selenium does not support overlapping elements detection
    on any command except `click`. Hence, when you call `click` on an element,
    and there is some overlay on top of it
    (e.g. for the sake of indicating "loading in progress"),
    that is going to disappear after some time,
    then Selenium will detect such overlap,
    that tells Selene to wait for it to disappear.
    But for any other command (double_click, context_click, type, etc.)
    Selenium will not and so Selene will not wait.
    Hence, if you want to wait in such cases, turn on this option.
    Just keep in mind, that it will work only for web tests, not mobile.
    """

    # TODO: better name? now technically it's not a decorator but decorator_builder...
    # or decorator_factory...
    # yet in python they call it just «decorator with args» or «decorator with params»
    # so technically we are correct naming it simply _wait_decorator
    # by type hint end users yet will see the real signature
    # and hence guess its «builder-like» nature
    # yet... should we for verbosity distinguish options
    # that a decorator factories from options that are simple decorators?
    # maybe better time to decide on this will be once we have more such options :p
    _wait_decorator: Callable[[Wait[E]], Callable[[F], F]] = lambda w: lambda f: f
    """
    Is used when performing any element command and assertion (i.e. should)
    Hence, can be used to log corresponding commands with waits,
    and integrate with something like allure reporting;)

    Yet prefixed with underscore, indicating that method is experimental,
    and so can be renamed or change its type signature, etc.

    The default value of this option just does nothing.

    See example of implementing and setting a custom wait decorator
    at https://github.com/yashaka/selene/blob/master/examples/log_all_selene_commands_with_wait.py

    Examples:
        And here is how you can use some predefined wait decorators in Selene,
        for example, to configure logging Selene commands to Allure report:

        >>> from selene.support.shared import browser
        >>> from selene import support
        >>> import allure_commons
        >>>
        >>> browser.config._wait_decorator = support._logging.wait_with(
        >>>   context=allure_commons._allure.StepContext
        >>> )
    """

    # TODO: why we name it as hook_* why not handle_* ?
    #       what would be proper style?
    hook_wait_failure: Optional[Callable[[TimeoutException], Exception]] = None
    '''
    A handler for all exceptions, thrown on failed waiting for timeout.
    Should process the original exception and rethrow it or the modified one.
    '''

    reports_folder: str = os.path.join(
        os.path.expanduser('~'),
        '.selene',
        'screenshots',
        str(round(time.time() * 1000)),
    )
    """
    A folder to save screenshots and page sources on failure.
    """
    save_screenshot_on_failure: bool = True
    """
    A flag to indicate whether to save screenshot on failure or not.
    If saved, will be also logged to the console on failure.
    """
    save_page_source_on_failure: bool = True
    """
    A flag to indicate whether to save page source on failure or not.
    If saved, will be also logged to the console on failure.
    """
    # TODO: consider making public
    _counter: itertools.count = itertools.count(start=int(round(time.time() * 1000)))
    """
    A counter, currently used for incrementing screenshot and page source names
    """
    last_screenshot: Optional[str] = None
    last_page_source: Optional[str] = None
    # TODO: is a _strategy suffix a good naming convention in this context?
    #       maybe yes, because we yet accept config in it...
    #       so we expect it to be a Strategy of some bigger Context
    _save_screenshot_strategy: Callable[
        [Config, Optional[str]], Any
    ] = lambda config, path=None: fp.thread(
        path,
        lambda path: (
            config._generate_filename(suffix='.png') if path is None else path
        ),
        lambda path: (
            os.path.join(path, f'{next(config._counter)}.png')
            if path and not path.lower().endswith('.png')
            else path
        ),
        fp.do(
            fp.pipe(
                os.path.dirname,
                lambda folder: (
                    os.makedirs(folder)
                    if folder and not os.path.exists(folder)
                    else ...
                ),
            )
        ),
        fp.do(
            lambda path: (
                warnings.warn(
                    "name used for saved screenshot does not match file "
                    "type. It should end with an `.png` extension",
                    UserWarning,
                )
                if not path.lower().endswith('.png')
                else ...
            )
        ),
        lambda path: (path if config.driver.get_screenshot_as_file(path) else None),
        fp.do(
            lambda path: setattr(config, 'last_screenshot', path)
        ),  # On refactor>rename, we may miss it here :( better would be like:
        #  setattr(config, config.__class__.last_screenshot.name, path)
        #  but currently .name will return '__boxed_last_screenshot' :(
        #  think on how we can resolve this...
    )
    """
    Defines a strategy for saving a screenshot.

    The default strategy saves a screenshot to a file,
    and stores the path to `config.last_screenshot`.
    """

    _save_page_source_strategy: Callable[
        [Config, Optional[str]], Any
    ] = lambda config, path=None: fp.thread(
        path,
        lambda path: (
            config._generate_filename(suffix='.html') if path is None else path
        ),
        lambda path: (
            os.path.join(path, f'{next(config._counter)}.html')
            if path and not path.lower().endswith('.html')
            else path
        ),
        fp.do(
            fp.pipe(
                os.path.dirname,
                lambda folder: (
                    os.makedirs(folder)
                    if folder and not os.path.exists(folder)
                    else ...
                ),
            )
        ),
        fp.do(
            lambda path: (
                warnings.warn(
                    "name used for saved page source does not match file "
                    "type. It should end with an `.html` extension",
                    UserWarning,
                )
                if not path.lower().endswith('.html')
                else ...
            )
        ),
        lambda path: (path, config.driver.page_source),
        fp.do(lambda path_and_source: fp.write_silently(*path_and_source)),
        lambda path_and_source: path_and_source[0],
        fp.do(
            lambda path: setattr(config, 'last_page_source', path)
        ),  # On refactor>rename, we may miss it here :( better would be like:
        #  setattr(config, config.__class__.last_screenshot.name, path)
        #  but currently .name will return '__boxed_last_screenshot' :(
        #  think on how we can resolve this...
    )
    """
    Defines a strategy for saving a page source on failure.

    The default strategy saves a page_source to a file,
    and stores the path to `config.last_page_source`.
    """

    # TODO: consider adding option to disable persistence of all not-overridden options
    #       or marking some of them as not persistent
    #       (i.e. unbind some of them keeping the previous value set)
    def with_(self, **options_to_override) -> Config:
        """

        Returns:
            A new config with overridden options that were specified as arguments.

                All other config options will be shallow-copied
                from the current config.
                Those other options that are of immutable types,
                like `int` - will be also copied by reference,
                i.e. in a truly shallow way.

        Parameters:
            **options_to_override:
                options to override in the new config.

                Technically "override" here means:
                "deep copy option storage and update its value to the specified one".
                All other option storages will be:
                "shallow copied from the current config".

                If `driver_name` is among `options_to_override`,
                and `driver` is not among them,
                and `self._override_driver_with_all_driver_like_options` is True,
                then `driver` will be implicitly added to the options to override,
                i.e. `with_(driver_name='firefox')` will be equivalent
                to `with_(driver_name='firefox', driver=...)`.
                The latter gives a readable and concise shortcut
                to spawn more than one browser:

                ```python
                from selene import Config

                config = Config(timeout=10.0, base_url='https://autotest.how')
                chrome = config.driver  # chrome is default browser
                firefox_config = config.with_(driver_name='firefox')
                firefox = firefox_config.driver
                edge_config = config.with_(driver_name='edge')
                edge = edge_config.driver
                ```

                Same logic applies to `remote_url`,
                and all other config.*driver* options.
        """
        options = (
            {'driver': ..., **options_to_override}
            if (
                self._override_driver_with_all_driver_like_options
                and 'driver' not in options_to_override
                and any('driver' in key for key in options_to_override)
            )
            else options_to_override
        )
        return persistent.replace(self, **options)

    def _generate_filename(self, prefix='', suffix=''):
        path = self.reports_folder
        next_id = next(self._counter)
        filename = f'{prefix}{next_id}{suffix}'
        file = os.path.join(path, f'{filename}')

        folder = os.path.dirname(file)
        if not os.path.exists(folder) and folder:
            os.makedirs(folder)

        return file

    # TODO: consider moving this injection to the WaitingEntity.wait method
    #       to build Wait object instead of config.wait
    def _inject_screenshot_and_page_source_pre_hooks(self, hook):
        # TODO: consider moving hooks to class methods accepting config as argument
        #       or refactor somehow to eliminate all times defining hook fns
        def save_and_log_screenshot(error: TimeoutException) -> Exception:
            path = self._save_screenshot_strategy(self)
            return TimeoutException(
                error.msg
                + f'''
Screenshot: file://{path}'''
            )

        def save_and_log_page_source(error: TimeoutException) -> Exception:
            filename = (
                # TODO: this dependency to last_page_source might lead to code,
                #       when wrong last_page_source name is taken
                self.last_screenshot.replace('.png', '.html')
                if self.last_screenshot
                else self._generate_filename(suffix='.html')
            )

            path = self._save_page_source_strategy(self, filename)
            return TimeoutException(error.msg + f'\nPageSource: file://{path}')

        return fp.pipe(
            save_and_log_screenshot if self.save_screenshot_on_failure else None,
            save_and_log_page_source if self.save_page_source_on_failure else None,
            hook,
        )

    # TODO: maybe here wait_factory would be better name?
    #       yes, it's also a strategy, but completely not connected with other
    #       driver lifecycle strategies
    _build_wait_strategy: Callable[
        [Config], Callable[[E], Wait[E]]
    ] = lambda config: lambda entity: Wait(
        entity,
        at_most=config.timeout,
        or_fail_with=config._inject_screenshot_and_page_source_pre_hooks(
            config.hook_wait_failure
        ),
        _decorator=config._wait_decorator,
    )
    """
    A strategy for building a Wait object based on other config options
    like `config.timeout`, `config.hook_wait_failure`, `config._wait_decorator`, etc.
    """

    # TODO: we definitely not need it inside something called Config,
    #       especially "base interface like config
    #       consider refactor to wait_factory as configurable config property
    # TODO: should we move it config._executor.wait(entity) ?
    def wait(self, entity: E) -> Wait[E]:
        return self._build_wait_strategy(self)(entity)



from functools import reduce
from typing import Tuple, ContextManager, Dict, Any, Iterable
from typing_extensions import Protocol

from selenium.webdriver import Keys


class _ContextManagerFactory(Protocol):
    def __call__(
        self, *, title: str, params: Dict[str, Any], **kwargs
    ) -> ContextManager:
        ...


class _default_translations:
    remove_verbosity = (
        ('browser.element', 'element'),
        ('browser.all', 'all'),
        ("'css selector', ", ""),
        ('((', '('),
        ('))', ')'),
    )
    identify_assertions = (
        (': has ', ': have '),
        (': have ', ': should have '),
        (': is ', ': should be '),
        (' and is ', ' and be '),
        (' and has ', ' and have '),
    )
    key_codes_to_names = [
        (f"({repr(value)},)", key)
        for key, value in Keys.__dict__.items()
        if not key.startswith('__')
    ]


def wait_with(
    *,
    context: _ContextManagerFactory,
    translations: Iterable[Tuple[str, str]] = (
        *_default_translations.remove_verbosity,
        *_default_translations.identify_assertions,
        *_default_translations.key_codes_to_names,
    ),
):
    """
    :return:
        Decorator factory to pass to Selene's config._wait_decorator
        for logging commands with waiting built in
    :param context:
        Allure-like ContextManager factory
        (i.e. a type/class or function to return python context manager),
        that builds a context manager based on title string and params dict
    :param translations:
        Iterable of translations as (from, to) substitution pairs
        to apply to final title string to log

    Examples:
        Given added allure dependency to your project,
        you can configure logging Selene commands to Allure report as simply as:

        >>> from selene import browser, support
        >>> import allure_commons
        >>>
        >>> browser.config._wait_decorator = support._logging.wait_with(
        >>>     context=allure_commons._allure.StepContext
        >>> )

        You also can pass a list of translations to be applied to final message to log,
        something like:

        >>> from selene import browser, support
        >>> import allure_commons
        >>>
        >>> browser.config._wait_decorator = support._logging.wait_with(
        >>>   context=allure_commons._allure.StepContext,
        >>>   translations=(
        >>>         ('browser.element', '$'),
        >>>         ('browser.all', '$$'),
        >>>   )
        >>> )

        You can also implement your own version of StepContext – feel free to use
        Alure's context manager as example.
        Or check some examples in Selene's tests, like this one:
        browser__config__wait_decorator_with_decorator_from_support_logging_test.py
    """

    def decorator_factory(wait):
        def decorator(for_):
            def decorated(fn):
                title = f'{wait.entity}: {fn}'

                def translate(initial: str, item: Tuple[str, str]):
                    old, new = item
                    return initial.replace(old, new)

                translated_title = reduce(
                    translate,
                    translations,
                    title,
                )

                with context(title=translated_title, params={}):
                    return for_(fn)

            return decorated

        return decorator

    return decorator_factory



# MIT License
#
# Copyright (c) 2015-2022 Iakiv Kramarenko
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import sys
from typing import Union, Optional

import typing
from selenium.webdriver import Keys

from selene.core.entity import Element, Collection, Browser
from selene.core.wait import Command
from selenium.webdriver import ActionChains
from selenium.webdriver.common.actions import interaction
from selenium.webdriver.common.actions.action_builder import ActionBuilder
from selenium.webdriver.common.actions.pointer_input import PointerInput


def save_screenshot(path: Optional[str] = None) -> Command[Browser]:
    command: Command[Browser] = Command(
        'save screenshot',
        lambda browser: browser.config._save_screenshot_strategy(browser.config, path),
    )

    if isinstance(path, Browser):
        # somebody passed command as `.perform(command.save_screenshot)`
        # not as `.perform(command.save_screenshot())`
        browser = path
        command.__call__(browser)

    return command


def save_page_source(path: Optional[str] = None) -> Command[Browser]:
    command: Command[Browser] = Command(
        'save page source',
        lambda browser: browser.config._save_page_source_strategy(browser.config, path),
    )

    if isinstance(path, Browser):
        # somebody passed command as `.perform(command.save_screenshot)`
        # not as `.perform(command.save_screenshot())`
        browser = path
        command.__call__(browser)

    return command


select_all: Command[Element] = Command(
    'select all by ctrl+a or cmd+a for mac',
    lambda element: typing.cast(Element, element)
    .locate()
    .send_keys(
        (Keys.COMMAND if sys.platform == 'darwin' else Keys.CONTROL) + 'a' + Keys.NULL,
    ),
)


# TODO: can we make it work for both mobile and web?
#       should we selectively choose proper interaction.POINTER_TOUCH below?
def _long_press(duration=1.0):
    def func(element: Element):
        located_element = element.locate()
        driver = element.config.driver
        actions: ActionChains = ActionChains(driver)

        actions.w3c_actions = ActionBuilder(
            driver, mouse=PointerInput(interaction.POINTER_TOUCH, 'touch')
        )
        (
            actions.w3c_actions.pointer_action.move_to(located_element)
            .pointer_down()
            .pause(duration)
            .release()
        )
        actions.perform()

    command = Command(f'long press with duration={duration}', func)

    if isinstance(duration, Element):
        # somebody passed command as `.perform(command.long_press)`
        # not as `.perform(command.long_press())`
        # TODO: refactor to really allow such use case without conflicts on types
        element = duration
        command.__call__(element)

    return command


class js:  # pylint: disable=invalid-name
    @staticmethod
    def set_value(value: Union[str, int]) -> Command[Element]:
        def func(element: Element):
            element.execute_script(
                """
                var text = arguments[0];
                var maxlength = element.getAttribute('maxlength') === null
                    ? -1
                    : parseInt(element.getAttribute('maxlength'));
                element.value = maxlength === -1
                    ? text
                    : text.length <= maxlength
                        ? text
                        : text.substring(0, maxlength);
                return null;
                """,
                str(value),
            )

        return Command(f'set value by js: {value}', func)

    @staticmethod
    def type(keys: Union[str, int]) -> Command[Element]:
        def func(element: Element):
            element.execute_script(
                """
                textToAppend = arguments[0];
                var value = element.value || '';
                var text = value + textToAppend;
                var maxlength = element.getAttribute('maxlength') === null
                    ? -1
                    : parseInt(element.getAttribute('maxlength'));
                element.value = maxlength === -1
                    ? text
                    : text.length <= maxlength
                        ? text
                        : text.substring(0, maxlength);
                return null;
                """,
                str(keys),
            )

        return Command(f'set value by js: {keys}', func)

    scroll_into_view: Command[Element] = Command(
        'scroll into view',
        lambda element: element.execute_script('element.scrollIntoView(true)'),
    )

    click: Command[Element] = Command(
        'click',
        # TODO: should we process collections too? i.e. click through all elements?
        lambda element: element.execute_script('element.click()'),
    )

    clear_local_storage: Command[Browser] = Command(
        'clear local storage',
        lambda browser: browser.driver.execute_script('window.localStorage.clear()'),
    )

    clear_session_storage: Command[Browser] = Command(
        'clear local storage',
        lambda browser: browser.driver.execute_script('window.sessionStorage.clear()'),
    )

    remove: Command[Union[Element, Collection]] = Command(
        'remove',
        lambda entity: (
            entity.execute_script('element.remove()')
            if not hasattr(entity, '__iter__')
            else [element.execute_script('element.remove()') for element in entity]
        ),
    )

    @staticmethod
    def set_style_property(name: str, value: Union[str, int]) -> Command[Element]:
        return Command(
            f'set element.style.{name}="{value}"',
            lambda entity: (
                entity.execute_script(f'element.style.{name}="{value}"')
                if not hasattr(entity, '__iter__')
                else [
                    element.execute_script(f'element.style.{name}="{value}"')
                    for element in entity
                ]
            ),
        )

    set_style_display_to_none: Command[Union[Element, Collection]] = Command(
        'set element.style.display="none"',
        lambda entity: (
            entity.execute_script('element.style.display="none"')
            if not hasattr(entity, '__iter__')
            else [
                element.execute_script('element.style.display="none"')
                for element in entity
            ]
        ),
    )

    set_style_display_to_block: Command[Union[Element, Collection]] = Command(
        'set element.style.display="block"',
        lambda entity: (
            entity.execute_script('element.style.display="block"')
            if not hasattr(entity, '__iter__')
            else [
                element.execute_script('element.style.display="block"')
                for element in entity
            ]
        ),
    )

    set_style_visibility_to_hidden: Command[Union[Element, Collection]] = Command(
        'set element.style.visibility="hidden"',
        lambda entity: (
            entity.execute_script('element.style.visibility="hidden"')
            if not hasattr(entity, '__iter__')
            else [
                element.execute_script('element.style.visibility="hidden"')
                for element in entity
            ]
        ),
    )

    set_style_visibility_to_visible: Command[Union[Element, Collection]] = Command(
        'set element.style.visibility="visible"',
        lambda entity: (
            entity.execute_script('element.style.visibility="visible"')
            if not hasattr(entity, '__iter__')
            else [
                element.execute_script('element.style.visibility="visible"')
                for element in entity
            ]
        ),
    )
